from __future__ import annotations

import getpass
from enum import Enum
from typing import List, Optional, Union

import yaml
from pydantic import FilePath, PrivateAttr

from osw.model.static import OswBaseModel


class CredentialManager(OswBaseModel):
    """Handles credentials"""

    cred_filepath: Optional[Union[Union[str, FilePath], List[Union[str, FilePath]]]]
    """Filepath to yaml file with credentials for osw and connected services"""
    cert_filepath: Optional[Union[Union[str, FilePath], List[Union[str, FilePath]]]]
    """Filepath to the certificates for osw and connected services"""

    _credentials: List[BaseCredential] = PrivateAttr([])
    """in memory credential store"""

    class BaseCredential(OswBaseModel):
        """Abstract base class for credentials"""

        iri: str
        """the iri the credential is valid for"""

    class UserPwdCredential(BaseCredential):
        """a username - password credential"""

        username: str
        """the user identifier"""
        password: str
        """the users password"""

    class OAuth1Credential(BaseCredential):
        """OAuth1 credentials. See https://requests-oauthlib.readthedocs.io/en/latest/oauth1_workflow.html"""

        consumer_token: str
        """consumer token"""
        consumer_secret: str
        """consumer secret """
        access_token: str
        """access token"""
        access_secret: str
        """access secret"""

    class CredentialFallback(Enum):
        """Modes of handling missing credentials

        Attributes
        ----------
        none:
            throw error
        ask:
            use getpass to ask for credentials
        """

        ask = "ask"  # use getpass to ask for credentials
        none = "none"  # throw error

    class CredentialConfig(OswBaseModel):
        """Reads credentials from a yaml file"""

        iri: str
        """internationalized resource identifier / address of the service, may contain protocol, domain, port and path
            matches by "contains" returning the shortest match"""
        fallback: Optional[CredentialManager.CredentialFallback] = "none"
        """The fallback strategy if no credential was found for the given origin"""

    def get_credential(self, config: CredentialConfig) -> BaseCredential:
        """Reads credentials from a yaml file or the in memory store

        Parameters
        ----------
        config:
            see CredentialConfig

        Returns
        -------
        credential :
            Credential, contain attributes 'username' and 'password' and the matching iri.
        """

        _file_credentials: List[CredentialManager.BaseCredential] = []
        if self.cred_filepath:
            filepaths = self.cred_filepath
            if type(filepaths) is not list:
                filepaths = [filepaths]

            for filepath in filepaths:
                if filepath != "":
                    with open(filepath, "r") as stream:
                        try:
                            accounts = yaml.safe_load(stream)
                            for iri in accounts.keys():
                                if (
                                    "username" in accounts[iri]
                                    and "password" in accounts[iri]
                                ):
                                    cred = CredentialManager.UserPwdCredential(
                                        username=accounts[iri]["username"],
                                        password=accounts[iri]["password"],
                                        iri=iri,
                                    )
                                    _file_credentials.append(cred)
                                if (
                                    "consumer_token" in accounts[iri]
                                    and "consumer_secret" in accounts[iri]
                                    and "access_token" in accounts[iri]
                                    and "access_secret" in accounts[iri]
                                ):
                                    cred = CredentialManager.OAuth1Credential(
                                        consumer_token=accounts[iri]["consumer_token"],
                                        consumer_secret=accounts[iri][
                                            "consumer_secret"
                                        ],
                                        access_token=accounts[iri]["access_token"],
                                        access_secret=accounts[iri]["access_secret"],
                                        iri=iri,
                                    )
                                    _file_credentials.append(cred)
                        except yaml.YAMLError as exc:
                            print(exc)

        match_iri = ""
        cred = None
        creds = _file_credentials + self._credentials
        for _cred in creds:
            iri = _cred.iri
            if config.iri in iri:
                if match_iri == "" or len(match_iri) > len(
                    iri
                ):  # use the less specific match
                    match_iri = iri
                    cred = _cred

        if cred is None:
            if config.fallback is CredentialManager.CredentialFallback.ask:
                print(
                    f"No credentials for {config.iri} found. Please use the prompt to login"
                )
                username = input("Enter username: ")
                password = getpass.getpass("Enter password: ")
                cred = CredentialManager.UserPwdCredential(
                    username=username, password=password, iri=config.iri
                )
        return cred

    def add_credential(self, cred: BaseCredential):
        """adds a credential to the in memory store

        Parameters
        ----------
        cred
            the credential to add
        """
        self._credentials.append(cred)


CredentialManager.CredentialConfig.update_forward_refs()



import copy
import getpass
from typing import Dict, List, Optional, Tuple, Union

import mwclient
import mwparserfromhell
import numpy as np
import yaml
from jsonpath_ng.ext import parse
from pydantic import FilePath

from osw.model.static import OswBaseModel
from osw.utils.util import parallelize


def read_domains_from_credentials_file(
    credentials_file_path: Union[str, FilePath]
) -> Tuple[List[str], Dict[str, Dict[str, str]]]:
    """Reads domains and credentials from a yaml file

    Parameters
    ----------
    credentials_file_path
        Path to the yaml file with the credentials

    Returns
    -------

    """
    with open(credentials_file_path, "r") as stream_:
        try:
            accounts_dict = yaml.safe_load(stream_)
            domains_list = list(accounts_dict.keys())
            if len(domains_list) == 0:
                raise ValueError("No domain found in accounts.pwd.yaml!")
            return domains_list, accounts_dict
        except yaml.YAMLError as exc_:
            print(exc_)


def read_credentials_from_yaml(
    password_file: Union[str, FilePath], domain: str = None
) -> dict:
    """Reads credentials from a yaml file

    Parameters
    ----------
    password_file :
        Path to the yaml file with the credentials.
    domain:
        Domain of the OSW instance, as specifed in the yaml file.

    Returns
    -------
    credentials :
        Dictionary with the credentials, expected to
        contain keys 'username' and 'password'.
    """
    if password_file != "":
        with open(password_file, "r") as stream:
            try:
                accounts = yaml.safe_load(stream)
                if domain is not None and domain in accounts.keys():
                    domain = domain
                elif len(accounts.keys()) > 0:
                    domain = list(accounts.keys())[0]
                    if len(accounts.keys()) > 0:
                        domain = list(accounts.keys())[0]
                user = accounts[domain]["username"]
                password = accounts[domain]["password"]
            except yaml.YAMLError as exc:
                print(exc)
    else:
        user = input("Enter bot username (username@botname)")
        password = getpass.getpass("Enter bot password")
    return {"username": user, "password": password}


def create_site_object(
    domain: str, password_file: Union[str, FilePath] = "", credentials: dict = None
) -> mwclient.client.Site:
    """
    Parameters
    ----------
    domain :
        Domain of the OSW instance, as specifed in the yaml file
    password_file :
        path to file with <username>\n<password>
    credentials :
        Dictionary with the credentials (username, password)
    Returns
    -------
    site : mwclient.client.Site
        Site object from mwclient lib
    """
    domain_dict = {
        "wiki-dev": {"Address": "wiki-dev.open-semantic-lab.org"},
        "onterface": {"Address": "onterface.open-semantic-lab.org:"},
    }
    if domain in domain_dict.keys():
        domain = domain_dict[domain]["Address"]

    site = mwclient.Site(domain, path="/w/")
    if credentials is None:
        credentials = read_credentials_from_yaml(password_file, domain)
    # else:
    #     credentials = credentials
    # Login with dictionary unpacking:
    # site.login(**credentials)
    # Explicit login:
    site.login(username=credentials["username"], password=credentials["password"])
    del credentials
    return site


class SearchParam(OswBaseModel):
    """Search parameters for semantic and prefix search"""

    query: Union[str, List[str]]
    parallel: Optional[bool] = False  # is set to true if query is a list longer than 5
    debug: Optional[bool] = True
    limit: Optional[int] = 1000

    # todo: @Simon: Bad style? Better to make it explicit in every function using it?
    def __init__(self, **data):
        super().__init__(**data)
        if isinstance(self.query, str):
            self.query = [self.query]
        if len(self.query) > 5:
            self.parallel = True


def prefix_search(
    site: mwclient.client.Site, text: Union[str, SearchParam]
) -> List[str]:
    """Standard query. Equivalent to the following mediawiki API call
    api.php?action=query&list=prefixsearch&pssearch=Star Wars

    Parameters
    ----------
    site :
        Site object from mwclient lib
    text :
        Query text or instance of SearchParam

    Returns
    -------
    page_list :
        List of page titles
    """
    if isinstance(text, str):
        query = SearchParam(query=[text])
    elif isinstance(text, list):
        query = SearchParam(query=text)
    else:
        query = text
    if len(query.query) > 5:
        query.parallel = True

    def prefix_search_(single_text):
        page_list = list()
        result = site.api(
            "query",
            list="prefixsearch",
            pssearch=single_text,
            pslimit=query.limit,
            format="json",
        )
        if len(result["query"]["prefixsearch"]) == 0:
            if query.debug:
                print("No results")
        else:
            for page in result["query"]["prefixsearch"]:
                title = page["title"]
                if query.debug:
                    print(title)
                page_list.append(title)
        return page_list

    if query.parallel:
        query_results = parallelize(
            func=prefix_search_, iterable=query.query, flush_at_end=query.debug
        )
    else:
        query_results = [prefix_search_(single_text=sq) for sq in query.query]

    return [item for sublist in query_results for item in sublist]
    # todo: @Simon: a list of lists of strings (sublist for each query in query list)
    #  or a list of strings (results of all queries combined)?
    #  The last option would not change the behavior of the function, but would
    # return page_list  # original return


def semantic_search(
    site: mwclient.client.Site, query: Union[str, List[str], SearchParam]
) -> List[str]:
    """Semantic query

    Parameters
    ----------
    site :
        Site object from mwclient lib
    query :
        (List of) query text(s) or instance of SearchParam

    Returns
    -------
    page_list:
        List of page titles
    """
    if isinstance(query, str):
        query = SearchParam(query=[query])
    elif isinstance(query, list):
        query = SearchParam(query=query)
    if len(query.query) > 5:
        query.parallel = True

    def semantic_search_(single_query):
        page_list = list()
        single_query += f"|limit={query.limit}"
        result = site.api("ask", query=single_query, format="json")
        if len(result["query"]["results"]) == 0:
            if query.debug:
                print("Query '{}' returned no results".format(single_query))
        else:
            if query.debug:
                print(
                    "Query '{}' returned {} results".format(
                        single_query, len(result["query"]["results"])
                    )
                )
            for page in result["query"]["results"].values():
                # why do we do the following?
                if "printouts" in page:
                    title = page["fulltext"]
                    if "#" not in title and query.debug:
                        print(title)
                        # original position of "page_list.append(title)" line
                    page_list.append(title)
        return page_list

    if query.parallel:
        query_results = parallelize(
            func=semantic_search_, iterable=query.query, flush_at_end=query.debug
        )
    else:
        query_results = [semantic_search_(single_query=sq) for sq in query.query]

    return [item for sublist in query_results for item in sublist]
    # todo: @Simon: a list of lists of strings (sublist for each query in query list)
    #  or a list of strings (results of all queries combined)?
    #  The last option would not change the behavior of the function, but would
    # return page_list  # original return


def search_wiki_page(title: str, site: mwclient.client.Site):
    """Page search wrapper that adds exact match functionality with ignore-case on
    top of the  prefix_search()'s functionality.

    Parameters
    ----------
    title :
        Title of the wiki page, e.g., User:Someone1234
    site :
        Site object from mwclient lib

    Returns
    -------
    result_dict : dict
    """
    result = site.api(
        "query", list="prefixsearch", pssearch=title, pslimit=1000, format="json"
    )
    if len(result["query"]["prefixsearch"]) == 0:
        return {"Result": False, "List": list()}
    else:
        exact_match = False
        page_title_list = list()
        for page in result["query"]["prefixsearch"]:
            page_title = page["title"]
            page_title_list.append(page_title)
            if page_title.lower() == title.lower():
                exact_match = True
        result_dict = {
            "Result": True,
            "List": page_title_list,
            "Exact match": exact_match,
        }
        return result_dict


def get_file_info_and_usage(
    site: mwclient.client.Site, title: Union[str, List[str], SearchParam]
) -> List[Dict[str, Union[Dict[str, str], List[str]]]]:
    """(For 'File' pages only) Get information about the file and its usage

    Parameters
    ----------
    site:
        Site object from mwclient lib.
    title:
        Title(s) of the wiki page(s) or instance of SearchParam.

    Returns
    -------
    result:
        Dictionary with page titles as keys and nested dictionary with keys 'info' and
        'usage'.

    Notes
    -----
    Query to reproduce:
        action=query
        format=json
        prop=imageinfo|fileusage
        titles=File%3AOSW857d85031d85425aa94db8b4720e84b7.png
        &iiprop=timestamp%7Cuser&fulimit=5000"

    Resources
    ---------
    Use the sandbox to design and test the queries:
    https://demo.open-semantic-lab.org/wiki/Special:ApiSandbox
    """
    if isinstance(title, str):
        query = SearchParam(query=[title], debug=False)
    elif isinstance(title, list):
        query = SearchParam(query=title, debug=False)
    else:  # SearchParam
        query = title
    if len(query.query) > 5:
        query.parallel = True

    def get_file_info_and_usage_(single_title):
        api_request_result = site.api(
            action="query",
            format="json",
            prop="imageinfo|fileusage",
            titles=single_title,
            iiprop="timestamp|user",
            fulimit=query.limit,
        )
        using_pages = []
        file_info = {
            "title": single_title,
            "author": "File not found or no creation logged",
            "timestamp": "File not found or no creation logged",
            "editor": [],
            "editing_timestamp": [],
        }

        if len(api_request_result["query"]["pages"]) == 0:
            if query.debug:
                print(f"Page not found: '{single_title}'!")
        else:
            image_info: List[Dict[str, str]] = []
            file_usage: List[Dict[str, Union[str, int]]] = []
            for page_id, page_dict in api_request_result["query"]["pages"].items():
                if page_dict["title"] == single_title:
                    image_info = page_dict.get("imageinfo", [])
                    file_usage = page_dict.get("fileusage", [])
            if len(image_info) != 0:
                file_info["author"] = image_info[0]["user"]
                file_info["timestamp"] = image_info[0]["timestamp"]
                for ii in image_info:
                    file_info["editor"].append(ii["user"])
                    file_info["editing_timestamp"].append(ii["timestamp"])
            if file_usage is not None:
                for fu_page_dict in file_usage:
                    using_pages.append(fu_page_dict["title"])
            if query.debug:
                # todo: find out why this message is printed (sometimes) when using the
                #  redirect,  which messes up the Progressbar
                #  printed messages do not appear in the MessageBuffer
                print(f"File info for '{single_title}' retrieved.")
        return {"info": file_info, "usage": using_pages}

    if query.parallel:
        api_request_results = parallelize(
            func=get_file_info_and_usage_,
            iterable=query.query,
            flush_at_end=query.debug,
        )
    else:
        api_request_results = [
            get_file_info_and_usage_(single_title=st) for st in query.query
        ]

    return api_request_results


def search_redirection_sources(
    site: mwclient.client.Site, target_title: str, debug: bool = False
):
    """Returns a list of pages redirecting to the page with target_title per #REDIRECT
    [[target]] syntax

    Parameters
    ----------
    site :
        Site object from mwclient lib
    target_title :
        Title of the target wiki page
    debug:
        Whether to print debugging messages

    Returns
    -------
    page_list : list of pages redirecting to the page with target_title
    """
    page_list = []
    result = site.api("query", titles=target_title, prop="redirects", format="json")
    if len(result["query"]["pages"]) == 0:
        if debug:
            print("No results")
    else:
        for page in result["query"]["pages"]:
            if "redirects" not in result["query"]["pages"][page]:
                if debug:
                    print("No results")
            else:
                for redirecting_source in result["query"]["pages"][page]["redirects"]:
                    title = redirecting_source["title"]
                    page_list.append(title)
    return page_list


def update_template_within_wikitext(
    text,
    template_text,
    delete=False,
    remove_empty_lines=False,
    overwrite_with_empty=False,
):
    """Updates the template parameters in an existing wiki <text> with a provided new
    <template_text>

    Parameters
    ----------
    text : str
        The existing text
    template_text : str
        The template text with new data
    delete : bool
        If true, params not defined in <template_text> get removed from <text>
    remove_empty_lines : bool
        If true, function will clean-up empty lines within the template code created by
        the underlying mwparserfromhell lib (wanted), but also within the wiki text
        around it (unwanted)
    overwrite_with_empty : bool
        If true, parameters in the existing template will be overwritten even if the
        parameter value in the template_text is empty

    Returns
    -------
    new_text : str
    """
    if template_text == text:
        return template_text
    if template_text == "":
        return text
    if text == "":
        return template_text
    new_code = mwparserfromhell.parse(template_text)
    new_template = new_code.filter_templates()[0]
    existing_code = mwparserfromhell.parse(text)
    template_names_matched = False
    for template in existing_code.filter_templates(recursive=True):
        if template.name.matches(new_template.name):
            existing_template = template
            template_names_matched = True
            break
    if template_names_matched:
        for p in new_template.params:
            if existing_template.has(p.name):
                if p.value != "" or overwrite_with_empty:
                    existing_template.get(p.name).value = p.value
            else:
                existing_template.add(p.name, p.value)
        if delete:
            for p in existing_template.params:
                if not new_template.has(p.name):
                    existing_template.remove(p)
    else:  # the original text did not contain a matching template
        # options:
        # 1) include the new template
        # 2) replace the existing template - but what if multiple tempaltes exist on
        #    that page?
        pass
    # print("Tmerged: \n" + str(existing_template))
    # print("Text merged: " + str(existing_code))
    new_text = str(existing_code)
    # this will cleanup empty lines within the template code (wanted), but also within
    # the wiki text around it (unwanted)
    if remove_empty_lines:
        new_text = "\n".join(
            [ll.rstrip() for ll in str(new_text).splitlines() if ll.strip()]
        )
    return new_text


def merge_wiki_page_text(
    text1, text2, template_name, subtemplate_param="", subtemplate_name=""
):
    """Not fully tested function!

    Parameters
    ----------
    text1
    text2
    template_name
    subtemplate_param
    subtemplate_name

    Returns
    -------
    non_empty_lines : str
    """

    if text1 == text2:
        return text1
    if text1 == "":
        return text2
    if text2 == "":
        return text1
    code1 = mwparserfromhell.parse(text1)
    code2 = mwparserfromhell.parse(text2)
    r1 = []
    r2 = []
    for template in code1.filter_templates(recursive=True):
        if template.name.matches(template_name):
            t1 = template
        if template.name.matches(subtemplate_name):
            r1.append(template)
    for template in code2.filter_templates(recursive=True):
        if template.name.matches(template_name):
            t2 = template
        if template.name.matches(subtemplate_name):
            r2.append(template)
    # print("T1: " + str(t1))
    # print("T2: " + str(t2))
    for p in t2.params:
        if not t1.has(p.name):
            t1.add(p.name, p.value)
    if subtemplate_param != "":
        t1.remove(subtemplate_param)
        for rel2 in r2:
            exists = False
            for rel1 in r1:
                all_params_equal = True
                for p in rel2.params:
                    if rel1.has(p.name):
                        if not rel1.get(p.name).value.matches(rel2.get(p.name).value):
                            all_params_equal = False
                        # print("Compare {}:{} with :{} -> res={}".format(p.name,
                        #       rel1.get(p.name).value,rel2.get(p.name).value,
                        #       rel1.get(p.name).value.matches(rel2.get(p.name).value)))
                    else:
                        all_params_equal = False
                if all_params_equal:
                    exists = True
            if not exists:
                r1.append(rel2)
        r1string = ""
        for rel1 in r1:
            r1string += "\r\n   " + str(rel1)
        t1.add(subtemplate_param, r1string)
    # print("Tmerged: " + str(t1))
    non_empty_lines = "\n".join(
        [ll.rstrip() for ll in str(t1).splitlines() if ll.strip()]
    )
    return non_empty_lines


def edit_wiki_page_with_content_merge(title, new_content, site, template_name):
    """Edits an existing wiki page, while merging the passed new content with the
    content of the existing page

    Parameters
    ----------
    title : str
        Title of the wiki page, e.g., User:Someone1234
    new_content : str
    site : mwclient.client.Site
        Site object from mwclient lib
    template_name : str

    Returns
    -------
    success : bool
    """
    search_result = search_wiki_page(title, site)
    if search_result["Result"]:
        source_page = site.pages[title]
        source_page_content = source_page.text()
        # todo: test function
        new_content = merge_wiki_page_text(
            new_content, source_page_content, template_name=template_name
        )
        target_page = site.pages[title]
        target_page.edit(new_content, "[bot] update of page content")
        success = True
    else:
        success = False
    return success


def create_flat_content_structure_from_wikitext(
    text: str, array_mode: str = "force"
) -> list:
    """Create a flat python dict (aka 'flat_content_structure' = 'wikiJson')
    representing the content of the page

    Parameters
    ----------
    text :
        the wiki source text
    array_mode :
        defines how to parse template params
        array_mode / param     value   value;  value;value     comment
        'force':               array   array   array           always create an array
        'only_multiple':       literal literal array           create only when more than one value is given
        'delimiter_present':   literal array   array           create array if at least one separator char is present

    Returns
    -------
    res :
        (aka 'flat_content_structure' = 'wikiJson')
    """

    res = []
    text = text.strip()
    existing_code = mwparserfromhell.parse(text)
    t_count = 0
    for i in range(0, len(existing_code.nodes)):
        n = existing_code.nodes.pop(
            0
        )  # returns first layer of nodes. filter() returns also template args
        if type(n) is mwparserfromhell.nodes.template.Template:
            t_count += 1
            wt = {}
            # print(f"Template: {t.name} = {t}")
            wt[str(n.name).strip()] = {}
            for p in n.params:
                # print(f"  Param: {p.name} = {p.value} ({type(p.value)})")
                wt[str(n.name).strip()][
                    str(p.name)
                ] = create_flat_content_structure_from_wikitext(
                    str(p.value), array_mode
                )
            res.append(wt)
        else:
            if len(res) == 0 or type(res[-1]) is dict:
                res.append("")
            res[-1] = res[-1] + str(n)  # append to previous string if exists
    # for i, x in enumerate(res):
    #    if type(x) is str: res[i] = x.strip() #remove whitespace
    res = [x for x in res if x and not (type(x) is str and x.isspace())]
    if t_count == 0:
        res = text
        values = str(text).strip().split(";")
        if array_mode == "force":
            res = values
        elif array_mode == "only_multiple" and len(values) > 1:
            res = values
        elif array_mode == "separator_present" and ";" in text:
            res = values
    return res


def get_wikitext_from_flat_content_dict(d: dict) -> str:
    """Create wiki source text from a flat python dict representing a wiki template

    Parameters
    ----------
    d :
        flat python dict
        e.g.: {"HeaderTemplate": {"param": "value"}}

    Returns
    -------
    wt :
        wiki text
    """
    wt = ""
    for key, value in d.items():
        # print("key: {}, valuetype: {}, value: {}".format(key, type(value), ""))
        if isinstance(value, dict):
            # print("dict")
            wt += "{{" + key
            wt += get_wikitext_from_flat_content_dict(value)
            wt += "\n}}"
        elif isinstance(value, list):
            # print("list")
            wt += "\n|{}=".format(key)
            string_index = 0
            for index, element in enumerate(value):
                if isinstance(element, dict):
                    wt += get_wikitext_from_flat_content_dict(element)
                    # wt += "\n{{" + element
                    # wt += get_wikitext_from_dict(element)
                    # wt += "\n}}"
                else:
                    if string_index != index:
                        print(
                            f"Warning: template param '{key}' has mixed template/"
                            f"string values: {value}"
                        )
                    if string_index > 0 and element and not element.strip().isspace():
                        wt += ";"
                    wt += element
                    string_index += 1
        else:
            # print("literal")
            wt += "\n|{}={}".format(key, value)
    return wt


def get_wikitext_from_flat_content_structure(content):
    """Create wiki source text from the content list (aka 'flat_content_structure' =
    'wikiJson') of the page

    Parameters
    ----------
    content : list
        content list, mixed objects (templates) and free text
        (aka 'flat_content_structure' = 'wikiJson')
        e.g.: [{"HeaderTemplate": {"param": "value"}}, "freetext",
               {"FooterTemplate": {"param2": "value"}}]

    Returns
    -------
    wt : wiki text
    """
    wt = ""
    for content_element in content:
        if isinstance(content_element, dict):
            wt += get_wikitext_from_flat_content_dict(content_element)
        elif isinstance(content_element, str):
            wt += content_element  # "\n" + content_element
        else:
            print(
                "Error: content element is not dict or string: {}".format(
                    content_element
                )
            )
    return wt


def wikiJson2SchemaJson(schema, wikiJson):
    """Create osl schema-compatible json from the content representation of a page
    (aka 'flat_content_structure' = 'wikiJson')

    Parameters
    ----------
    wikiJson : list
        content list, mixed objects (templates) and free text
        (aka 'flat_content_structure' = 'wikiJson')
        e.g.: [{"HeaderTemplate": {"param": "value"}}, "freetext",
               {"FooterTemplate": {"param2": "value"}}]

    Returns
    -------
    schemaJson : dict
        schema-compatible json
        e.g.: {"osl_template": "HeaderTemplate", "param": "value",
               "osl_wikitext": "freetext",
               "osl_footer": {"osl_template": "FooterTemplate",
                              "param2": "value2"}}
    """
    schemaJson = {}
    if (
        not isinstance(wikiJson[0], dict)
        or not isinstance(wikiJson[1], str)
        or not isinstance(wikiJson[2], dict)
    ):
        print("Error: Invalid wikiJson:", wikiJson)
        return schemaJson

    schemaJson = wikiJson2SchemaJsonRecursion(schema, wikiJson[0], wikiJson[2])
    schemaJson["osl_wikitext"] = wikiJson[1]
    return schemaJson


def wikiJson2SchemaJsonRecursion(schema: dict, wikiJson: dict, footerWikiJson=None):
    """internal recursion function of wikiJson2SchemaJson()

    Parameters
    ----------
    wikiJson : list
        wiki template representation
        e.g.: {"HeaderTemplate": {"param": "value"}}

    footerWikiJson : list
        wiki footer template representation
        e.g.: {"FooterTemplate": {"param2": "value2"}}

    Returns
    -------
    schemaJson : dict
        schema-compatible json
        e.g.: {"osl_template": "HeaderTemplate", "param": "value",
               "osl_footer": {"osl_template": "FooterTemplate", "param2": "value2"}}
    """
    schemaJson = {}
    if footerWikiJson is not None:
        schemaJson["osl_footer"] = wikiJson2SchemaJsonRecursion(schema, footerWikiJson)

    for key in wikiJson:
        value = wikiJson[key]
        if isinstance(value, list):
            schemaJson[key] = []
            for index, element in enumerate(value):
                element = value[index]
                if isinstance(element, dict):
                    if key == "extensions":
                        if (
                            footerWikiJson is not None
                        ):  # we asume that every extension provides also a footer template
                            nextFooter = footerWikiJson[
                                schemaJson["osl_footer"]["osl_template"]
                            ]["extensions"][index]
                            schemaJson[key].append(
                                wikiJson2SchemaJsonRecursion(
                                    schema, element, nextFooter
                                )
                            )
                    else:
                        schemaJson[key].append(
                            wikiJson2SchemaJsonRecursion(schema, element)
                        )
                else:
                    schemaJson[key].append(element)

        elif isinstance(value, dict):
            schemaJson = wikiJson2SchemaJsonRecursion(schema, value, footerWikiJson)
            schemaJson["osl_template"] = key
        else:
            schemaJson[key] = value

    if "osl_template" in schemaJson:
        jsonpath_expr = parse(
            "$..properties"
        )  # [?osl_template.default = "' + schemaJson['osl_template'] + '"]')
        schema_def = {}
        # print("Search for ", schemaJson['osl_template'])
        for match in jsonpath_expr.find(schema):
            value = match.value
            if (
                "osl_template" in value
                and "osl_template" in schemaJson
                and value["osl_template"]["default"] == schemaJson["osl_template"]
                or "osl_schema" in value
                and "osl_schema" in schemaJson
                and value["osl_schema"]["default"] == schemaJson["osl_schema"]
            ):
                schema_def = match.value  # ToDo: Resolve allOf...
                # print(schema_def)

        for key in list(schemaJson.keys()):
            if key in schema_def:
                if "type" in schema_def[key]:  # use schema to resolve types
                    if schema_def[key]["type"] == "integer":
                        schemaJson[key] = int(schemaJson[key])
                    elif schema_def[key]["type"] == "float":
                        schemaJson[key] = float(schemaJson[key])
                    elif schema_def[key]["type"] == "number":
                        schemaJson[key] = float(schemaJson[key])
                    elif schema_def[key]["type"] == "string":
                        schemaJson[key] = str(schemaJson[key])
                    elif schema_def[key]["type"] == "array":
                        if not isinstance(schemaJson[key], list):
                            del schemaJson[key]
                        elif len(schemaJson[key]) == 0:
                            del schemaJson[key]
                else:  # assume type==object
                    if isinstance(schemaJson[key], list) and len(schemaJson[key]) == 1:
                        schemaJson[key] = schemaJson[key][0]
            else:  # fall back
                if schemaJson[key] == "" and key == "extensions":
                    del schemaJson[key]
                elif isinstance(
                    schemaJson[key], list
                ):  # wikiJson defaults are lists, even for single or empty values
                    if len(schemaJson[key]) == 0:
                        del schemaJson[key]
                    # elif len(schemaJson[key]) == 1: schemaJson[key] = schemaJson[key][0]

    return schemaJson


def schemaJson2WikiJson(schemaJson, isRoot=True):
    """Create content representation of a page
    (aka 'flat_content_structure' = 'wikiJson') from the osl schema-compatible json

    Parameters
    ----------
    schemaJson : dict
        schema-compatible json
        e.g.: {"osl_template": "HeaderTemplate", "param": "value",
               "osl_wikitext": "freetext",
               "osl_footer": {"osl_template": "FooterTemplate", "param2": "value2"}}

    isRoot: boolean
        indicates first call in recursion

    Returns
    -------
    wikiJson : list
        content list, mixed objects (templates) and free text
        (aka 'flat_content_structure' = 'wikiJson')
        e.g.: [{"HeaderTemplate": {"param": "value"}}, "freetext",
               {"FooterTemplate": {"param2": "value"}}]
    """
    wikiJson = [{}, "", {}]
    # header, freetext, footer
    template = ""
    footer_template = ""
    if "osl_template" in schemaJson:
        template = schemaJson["osl_template"]
        wikiJson[0][template] = {}
    else:
        print(
            "Error: Mandatory property 'osl_template' not found in schemaJson",
            schemaJson,
        )
        return

    if "osl_wikitext" in schemaJson:
        wikiJson[1] = schemaJson["osl_wikitext"]
    if "osl_footer" in schemaJson:
        wikiJson[2] = schemaJson2WikiJson(schemaJson["osl_footer"], False)[0]
        footer_template = schemaJson["osl_footer"]["osl_template"]
        wikiJson[2][footer_template]["extensions"] = []

    for key in schemaJson:
        if (
            key.startswith("_")
            or key.startswith("osl_template")
            or key.startswith("osl_wikitext")
            or key.startswith("osl_footer")
        ):
            continue  # exclude private and reserved keywords
        if schemaJson[key] is None:
            continue
        if isinstance(schemaJson[key], list):
            wikiJson[0][template][key] = []
            for subSchemaJson in schemaJson[key]:
                subWikiJson = schemaJson2WikiJson(subSchemaJson, False)
                wikiJson[0][template][key].append(subWikiJson[0])
                if key == "extensions":
                    wikiJson[2][footer_template]["extensions"].append(subWikiJson[2])

        elif isinstance(schemaJson[key], dict):
            subWikiJson = schemaJson2WikiJson(schemaJson[key], False)
            wikiJson[0][template][key] = [subWikiJson[0]]
        else:
            wikiJson[0][template][key] = schemaJson[key]

    return wikiJson


def create_or_overwrite_wiki_page(title, content, site):
    """Creates a page with the passed title and content. If the page already exists,
    the prior content is replaced with the passed content.

    Parameters
    ----------
    title : str
        Title of the wiki page, e.g., User:Someone1234
    content : str
    site : mwclient.client.Site
        Site object from mwclient lib

    Returns
    -------
    success : bool
    """
    target_page = site.pages[title]
    target_page.edit(content, "[bot] create page")
    success = True
    return success


def delete_wiki_page(title, site, reason):
    """Deletes the wiki page with the passed title, if it was found (exact match!),
    otherwise returns False

    Parameters
    ----------
    title : str
        Title of the wiki page, e.g., User:Someone1234
    site : mwclient.client.Site
        Site object from mwclient lib
    reason : str

    Returns
    -------
    success : bool
    """
    try:
        page = site.pages[title]
        page.delete(reason=reason, watch=False, unwatch=True, oldimage=False)
        success = True
    except mwclient.errors.APIError:
        success = False
    return success


def create_or_update_wiki_page_with_template(
    title, content, site, overwrite_with_empty=False
):
    """Creates a wiki page with a template included in the content. If the page does
    already exist, the parameters within the template are update

    Parameters
    ----------
    title : str
        Title of the wiki page, e.g., User:Someone1234
    content : str
    site : mwclient.client.Site
        Site object from mwclient lib
    overwrite_with_empty : bool
        Decided whether a template parameter's value in an preexisting page is
        overwritten with an empty value

    Returns
    -------
    success : bool
    """
    search_result = search_wiki_page(title, site)
    if search_result["Result"]:
        existing_page = site.pages[title]
        existing_text = existing_page.text()
        # update the page's content (template only)
        updated_content = update_template_within_wikitext(
            text=existing_text,
            template_text=content,
            overwrite_with_empty=overwrite_with_empty,
        )
        success = create_or_overwrite_wiki_page(title, updated_content, site)
    else:
        # just create the page
        success = create_or_overwrite_wiki_page(title, content, site)
    return success


def find_dependencies(wikitext, debug=False):
    """Finds templates, properties and categories within wikitext

    Parameters
    ----------
    wikitext : str
    debug : bool
        Whether to print debugging messages

    Returns
    -------
    filtered_dependencies : list
    """
    dependencies = []
    code = mwparserfromhell.parse(wikitext)
    for template in code.filter_templates(recursive=True):
        if template.name.split(":")[0].isupper():
            if debug:
                print("MagicWord: {}".format(template.name))
        elif template.name[0] == "#":
            if debug:
                print("ParserFunction: {}".format(template.name))
            if "#set:" in template.name or "#declare:" in template.name:
                if (
                    "=" in template.name.split(":")[1]
                ):  # in case of '{{#set:HasIdPostfix={{{id_postfix}}} }}'
                    property_ = "Property:" + template.name.split(":")[1].split("=")[0]
                    dependencies.append(property_)
                    if debug:
                        print("=> {}".format(property_))
                for param in template.params:
                    property_ = "Property:" + param.split("=")[0]
                    dependencies.append(property_)
                    if debug:
                        print("=> {}".format(property_))
        else:
            if debug:
                print("Template: {}".format(template.name))
            template_name = str(template.name)
            if ":" not in template.name:
                template_name = "Template:" + template_name
            dependencies.append(template_name)
            if debug:
                print("=> {}".format(template_name))
    # for tag in code.filter_tags(recursive=True):
    #    if (debug): print("Tag: {}".format(tag))
    for link in code.filter_wikilinks(recursive=True):
        if "::" in link:
            if debug:
                print("Annotation: {}".format(link))
            property_ = "Property:" + link.split("::")[0].split("[[")[-1]
            dependencies.append(property_)
            if debug:
                print("=> {}".format(property_))
        if "Category:" in link:
            if debug:
                print("Category: {}".format(link))
            category = link.replace("[[", "").replace("]]", "")
            dependencies.append(str(category))
            if debug:
                print("=> {}".format(category))
        else:
            if debug:
                print("Link: {}".format(link))
    dependencies = np.unique(dependencies).tolist()  # remove duplicates
    filtered_dependencies = []  # do not manipulate the iterated object
    for dependency in dependencies:
        # ensure no leading or trailing white spaces
        dependency = dependency.strip()
        # very frew page titles listed here contain invalid characters "\n"
        dependency = dependency.replace("\n", "")
        # see https://www.semantic-mediawiki.org/wiki/Help:Special_properties
        if "Property:" in dependency and (" " in dependency or "_" in dependency):
            if debug:
                print(
                    "Info: Remove presumptive built-in property {}".format(dependency)
                )
        else:
            filtered_dependencies.append(dependency)
    return filtered_dependencies


def find_dependencies_recursively(title, site, dependencies=None, debug=False):
    """Finds all included templates, properties and categories within a wiki page

    Parameters
    ----------
    title : str
        Title of the wiki page, e.g., User:Someone1234
    site : mwclient.client.Site
        Site object from mwclient lib
    dependencies
        List of already found dependent titles
    debug : bool
        Whether to print debugging messages

    Returns
    -------
    dependencies : list
    """
    if dependencies is None:
        dependencies = []
    title = title.replace("{", "").replace("}", "")
    page = site.pages[title]
    content = page.text()
    new_dependencies = find_dependencies(content, debug)
    for dependency in new_dependencies:
        if dependency not in dependencies:  # for circular dependencies
            dependencies.append(dependency)
            if debug:
                print("Scan nested {}".format(dependency))
            find_dependencies_recursively(
                dependency, site, dependencies=dependencies, debug=debug
            )
    return dependencies


def extend_page_list_with_recursive_dependencies(page_list, site):
    """Finds all included templates, properties and categories for a list of wiki pages
    and appends them to the list

    Parameters
    ----------
    page_list : list or iterable
        List of pages
    site : mwclient.client.Site
        Site object from mwclient lib

    Returns
    -------
    page_list : list
    """
    page_list_iter = copy.deepcopy(page_list)
    for page in page_list_iter:
        # we provide the updates page list to prevent a revisit of the pages
        dependencies = find_dependencies_recursively(
            page, site, dependencies=page_list, debug=False
        )
        page_list.extend(dependencies)
    page_list = np.unique(page_list)
    return page_list


def copy_wiki_page(title0, title1, site0, site1, overwrite=True):
    """

    Parameters
    ----------
    title0 : str
        Title of the source page
    title1 : str
        Title of the target page
    site0 : mwclient.client.Site
        Source site object from mwclient lib
    site1 : mwclient.client.Site
        Target site object from mwclient lib
    overwrite : bool
        Whether to overwrite existing pages at target site

    Returns
    -------
    success: bool

    """
    if title0.lower() == title1.lower() and site0 == site1:
        # copy on it self = no action necessary
        success = True
    else:
        page0 = site0.pages[title0]
        content = page0.text()
        if overwrite:
            success = create_or_overwrite_wiki_page(title1, content, site1)
        else:
            search_result = search_wiki_page(title1, site1)
            if (
                search_result["Result"] and search_result["Exact match"]
            ):  # page already exists
                success = False
            else:  # search_result["Result"] == True/False, search_result["Exact match"] == False
                success = create_or_overwrite_wiki_page(title1, content, site1)
    return success


def copy_list_of_wiki_pages(title_list, site0, site1, overwrite, callback=None):
    """

    Parameters
    ----------
    title_list : list
    site0 : mwclient.client.Site
        Source site object from mwclient lib
    site1 : mwclient.client.Site
        Target site object from mwclient lib
    overwrite : bool
        Whether to overwrite existing pages at target site
    callback : NoneType or function
        Function passed over, to perform operation on the titles of the source pages and
        to be passed as title of the
        target pages. See examples below.
        Example functions:
            capitalize = lambda x: x.capitalize()
            def change_namespace(title, namespace):
                if ":" in namespace:
                    namespace = namespace.split(":")[0]
                if ":" in title:
                    splits = title.split(":")
                    old_name = splits[1].capitalize()
                    new_title = namespace + ":" + old_name
                else:
                    new_title = namespace + ":" + title.capitalize()
                return new_title
        Examples of passing a function as parameter:
            callback = capitalize
            callback = lambda x: x.lower()

    Returns
    -------
    results_dict : dict
        Dictionary, containing the results of the copying operations
    """

    success_list = list()
    fail_list = list()
    for title0 in title_list:
        if callback is None:
            title1 = title0
        else:
            title1 = callback(title0)
        success = copy_wiki_page(title0, title1, site0, site1, overwrite)
        if success:
            success_list.append(title1)
        else:
            fail_list.append(title1)
    results_dict = {
        "Successfully copied pages": success_list,
        "Pages failed to copy": fail_list,
    }
    return results_dict



import json

from osw.utils.templates import eval_handlebars_template


def test_category_template():
    template = """
{
    "@context": [
        {{#each subclass_of}}
        "/wiki/{{{.}}}?action=raw\u0026slot=jsonschema"{{#unless @last}},{{/unless}}
        {{/each}}
    ],
    "allOf": [
        {
            {{#each subclass_of}}
            "$ref": "/wiki/{{{.}}}?action=raw\u0026slot=jsonschema"{{#unless @last}},{{/unless}}
            {{/each}}
        }
    ],
    "type": "object",
    "uuid": "{{{uuid}}}",
    "title": "{{{name}}}",
    "title*": {
        {{#each label}}
        "{{{lang}}}": "{{{text}}}"{{#unless @last}},{{/unless}}
        {{/each}}
    },
    "description": "{{{description.[0].text}}}",
    "description*": {
        {{#each description}}
        "{{{lang}}}": "{{{text}}}"{{#unless @last}},{{/unless}}
        {{/each}}
    },
    "required": ["type"],
    "properties": {
        "type": {
            "default": ["{{{_page_title}}}"]
        }
    }
}"""

    data = {
        "type": ["Category:Category"],
        "subclass_of": ["Category:Category"],
        "uuid": "379d5a15-89c7-4c82-bc0d-e47938264d00",
        "label": [{"text": "OwlThing", "lang": "en"}],
        "metaclass": ["Category:OSW725a3cf5458f4daea86615fcbd0029f8"],
        "description": [
            {
                "text": "Represents the set of all individuals. In the DL literature this is often called the top concept.",
                "lang": "en",
            }
        ],
        "name": "OwlThing",
    }

    expected = """
{
  "@context": [
    "/wiki/Category:Category?action=raw&slot=jsonschema"
  ],
  "allOf": [
    {
      "$ref": "/wiki/Category:Category?action=raw&slot=jsonschema"
    }
  ],
  "type": "object",
  "uuid": "379d5a15-89c7-4c82-bc0d-e47938264d00",
  "title": "OwlThing",
  "title*": {
    "en": "OwlThing"
  },
  "description": "Represents the set of all individuals. In the DL literature this is often called the top concept.",
  "description*": {
    "en": "Represents the set of all individuals. In the DL literature this is often called the top concept."
  },
  "required": [
    "type"
  ],
  "properties": {
    "type": {
      "default": [
        "Category:OSW379d5a1589c74c82bc0de47938264d00"
      ]
    }
  }
}
"""

    output = json.loads(
        eval_handlebars_template(
            template,
            data,
            {"_page_title": "Category:OSW379d5a1589c74c82bc0de47938264d00"},
        )
    )

    assert output == json.loads(expected)


def test_metamodel_template():
    template = """
    {
        "title": "{{{name}}}",
        "required": ["uuid"],
        "properties": {
            "uuid": {
                "type": "string",
                "format": "uuid",
                "options": {
                    "hidden": true
                }
            } {{#each parameters}},
            "{{{name}}}": {
                "title": "{{{name}}}",
                "type": "{{{type}}}" {{#if default}},
                "default": "{{{default}}}"{{/if}}
            } {{/each}}{{#each submodels}},
            "{{{name}}}":
                {{> self}}
            {{/each}}
        }
    }"""

    data = {
        "submodels": [
            {
                "name": "Geometrie",
                "parameters": [{"name": "FaceArea"}, {"name": "dimension"}],
                "submodels": [
                    {
                        "name": "NegativeElectrode",
                        "parameters": [],
                        "submodels": [
                            {
                                "name": "ActiveMaterial",
                                "parameters": [{"name": "thickness"}],
                                "submodels": [],
                            }
                        ],
                    },
                    {
                        "name": "PositiveElectrode",
                        "parameters": [],
                        "submodels": [
                            {
                                "name": "ActiveMaterial",
                                "parameters": [{"name": "thickness"}],
                                "submodels": [],
                            }
                        ],
                    },
                ],
            }
        ],
        "type": ["Category:OSWecff4345b4b049218f8d6628dc2f2f21"],
        "uuid": "dab254dd-1006-4ddf-9554-64b7a5baf332",
        "name": "BattmoModel",
        "label": [{"text": "BattMo Model", "lang": "en"}],
    }

    expected = """{
        "title":"BattmoModel",
        "required":[
            "uuid"
        ],
        "properties":{
            "uuid":{
                "type":"string",
                "format":"uuid",
                "options":{
                    "hidden":true
                }
            },
            "Geometrie":{
                "title":"Geometrie",
                "required":[
                    "uuid"
                ],
                "properties":{
                    "uuid":{
                    "type":"string",
                    "format":"uuid",
                    "options":{
                        "hidden":true
                    }
                    },
                    "FaceArea":{
                    "title":"FaceArea",
                    "type":""
                    },
                    "dimension":{
                    "title":"dimension",
                    "type":""
                    },
                    "NegativeElectrode":{
                    "title":"NegativeElectrode",
                    "required":[
                        "uuid"
                    ],
                    "properties":{
                        "uuid":{
                            "type":"string",
                            "format":"uuid",
                            "options":{
                                "hidden":true
                            }
                        },
                        "ActiveMaterial":{
                            "title":"ActiveMaterial",
                            "required":[
                                "uuid"
                            ],
                            "properties":{
                                "uuid":{
                                "type":"string",
                                "format":"uuid",
                                "options":{
                                    "hidden":true
                                }
                                },
                                "thickness":{
                                "title":"thickness",
                                "type":""
                                }
                            }
                        }
                    }
                    },
                    "PositiveElectrode":{
                    "title":"PositiveElectrode",
                    "required":[
                        "uuid"
                    ],
                    "properties":{
                        "uuid":{
                            "type":"string",
                            "format":"uuid",
                            "options":{
                                "hidden":true
                            }
                        },
                        "ActiveMaterial":{
                            "title":"ActiveMaterial",
                            "required":[
                                "uuid"
                            ],
                            "properties":{
                                "uuid":{
                                "type":"string",
                                "format":"uuid",
                                "options":{
                                    "hidden":true
                                }
                                },
                                "thickness":{
                                "title":"thickness",
                                "type":""
                                }
                            }
                        }
                    }
                    }
                }
            }
        }
        }"""

    output = json.loads(eval_handlebars_template(template, data))

    assert output == json.loads(expected)



from uuid import UUID

from osw.core import OSW


def test_osw_id_to_uuid():
    osw_id = "OSW2ea5b605c91f4e5a95593dff79fdd4a5"
    expected = UUID("2ea5b605-c91f-4e5a-9559-3dff79fdd4a5")
    result = OSW.get_uuid(osw_id)
    assert result == expected


def test_uuid_to_osw_id():
    uuid = UUID("2ea5b605-c91f-4e5a-9559-3dff79fdd4a5")
    expected = "OSW2ea5b605c91f4e5a95593dff79fdd4a5"
    result = OSW.get_osw_id(uuid)
    assert result == expected



import os
from importlib import reload

import osw.model.entity as model
from osw.core import OSW
from osw.wtsite import WtSite

# create/update the password file under examples/accounts.pwd.yaml
pwd_file_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "accounts.pwd.yaml"
)
# pwd_file_path = "./accounts.pwd.yaml"
wtsite = WtSite.from_domain("wiki-dev.open-semantic-lab.org", pwd_file_path)
osw = OSW(site=wtsite)

# osw.fetch_schema() #this will load the current entity schema from the OSW instance.
# You may have to re-run the script to get the updated schema extension.
# Requires 'pip install datamodel-code-generator'
list_of_categories = [
    "Category:Entity",
    "Category:Item",
    # "Category:OSWd9aa0bca9b0040d8af6f5c091bf9eec7",  # User
    # "Category:OSWb8b6278763d54b0784eea9d3b3d183a4",  # Group
    # "Category:OSW3d238d05316e45a4ac95a11d7b24e36b",  # Location
    # "Category:OSW473d7a1ed48544d1be83b258b5810948",  # Site
    # "Category:OSW4bcd4a99a73f482ea40ac4210dfab836",  # Building
    # "Category:OSW6c4212f1a39342be963d2b9efd19c5c2",  # Floor
    # "Category:OSWc5ed0ed1e33c4b31887c67af25a610c1",  # Room
]
for i, cat in enumerate(list_of_categories):
    mode = "append"
    if i == 0:
        mode = "replace"
    osw.fetch_schema(OSW.FetchSchemaParam(schema_title=cat, mode=mode))
reload(model)



import copy
import re
import uuid as uuid_module
import warnings
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import deepl
import numpy as np
from geopy import Nominatim
from jsonpath_ng import ext as jp

import osw.data.mining as dm
from osw import wiki_tools as wt
from osw.auth import CredentialManager
from osw.core import OSW
from osw.data.mining import RegExPatternExtended
from osw.model import entity as model
from osw.wtsite import WtSite

# Constants
PACKAGE_ROOT_PATH = Path(__file__).parents[2]
CREDENTIALS_FILE_PATH_DEFAULT = PACKAGE_ROOT_PATH / "examples" / "accounts.pwd.yaml"
ENABLE_SORTING = True
REGEX_PATTERN: Dict[str, Union[str, Dict[str, str]]] = {
    "SAP OU number and name from DN": {
        "Pattern": r"CN=(.+)([0-9]{10})-(.+),OU=Abteilungen",
        "Groups": {2: "SAP OU number", 3: "SAP OU name"},
    },
    "Location name from DN": {
        "Pattern": r"CN=[A-Za-z]+-(\d+)_L_([^_]+),OU=Standorte",
        "Groups": {1: "SAP institute number", 2: "Location name"},
    },
    "Location/Site parts from DN": {
        "Pattern": r"CN=[A-Za-z]+-(\d+)_L_(([^_^ ^-]+)-([^_^ ]+) (\d+)),OU=Standorte",
        "Groups": {
            1: "SAP institute number",
            2: "Site name",
            3: "City",
            4: "Street",
            5: "House number",
        },
    },
    "UUID from full page title": {
        "Pattern": r"([A-Za-z]+):([A-Z]+)([a-z\d\-]+)",
        "Groups": {1: "Namespace", 2: "Prefix", 3: "UUID"},
    },
}
REGEX_PATTERN_LIST = [
    RegExPatternExtended(
        description="SAP OU number and name from DN",
        pattern=r"CN=(.+)([0-9]{10})-(.+),OU=Abteilungen",
        group_keys=["Something", "SAP OU number", "SAP OU name"],
    ),
    RegExPatternExtended(
        description="Location name from DN",
        pattern=r"CN=[A-Za-z]+\-(\d+)_L_([^_]+),OU=Standorte",
        group_keys=["SAP institute number", "Location name"],
    ),
    RegExPatternExtended(
        description="Location/Site parts from DN",
        pattern=r"CN=[A-Za-z]+\-(\d+)_L_(([^_^ ^-]+)-([^_^ ]+) (\d+))," r"OU=Standorte",
        group_keys=[
            "SAP institute number",
            "Site name",
            "City",
            "Street",
            "House number",
        ],
    ),
    RegExPatternExtended(
        description="UUID from full page title",
        pattern=r"([A-Za-z]+):([A-Z]+)([a-z\d\-]+)",
        group_keys=["Namespace", "Prefix", "UUID"],
    ),
]
REGEX_PATTERN = {rep.description: rep.dict() for rep in REGEX_PATTERN_LIST}
REGEX_PATTERN_LIB = {rep.description: rep for rep in REGEX_PATTERN_LIST}


# Classes
class HelperModel(model.OswBaseModel):
    # Custom attributes
    attributes_transformed: bool = False
    references_transformed: bool = False
    casted_instance: Any = None
    full_page_title: Optional[str]

    def transform_attributes(self, dd: dict) -> bool:
        if not self.attributes_transformed:
            uuid = uuid_module.uuid4()
            if hasattr(self, "uuid"):
                if self.uuid is not None:
                    uuid = self.uuid
            self.full_page_title = uuid_to_full_page_title(uuid=uuid)
            # All set successfully
            self.attributes_transformed = True
        return True

    def transform_references(self, dd: dict) -> bool:
        if not self.attributes_transformed:
            self.transform_attributes(dd)
        if not self.references_transformed:
            # Test casting
            superclass = self.__class__.__bases__[0]
            self.casted_instance = self.cast_none_to_default(cls=superclass)
            # All set successfully
            self.references_transformed = True
        return True

    def cast_to_superclass(self, dd):
        if not self.references_transformed:
            self.transform_references(dd)
        else:
            superclass = self.__class__.__bases__[0]
            self.casted_instance = self.cast_none_to_default(cls=superclass)
        return True


# Functions
def transform_attributes_and_merge(
    ent: dict, ent_as_dict: dict, sel_cls, inplace: bool = True
) -> dict:
    """Merge entities of a selected class within a dictionary of entities

    Parameters
    ----------
    ent
        Dictionary of entities
    ent_as_dict
        Dictionary of entities as dictionary
    sel_cls
        Selected class
    inplace
        Whether to transform_attributes the entities inplace

    Returns
    -------

    """
    if not inplace:
        ent = copy.deepcopy(ent)
        ent_as_dict = copy.deepcopy(ent_as_dict)
    ent, ent_as_dict = loop_and_call_method(
        entities=ent,
        method_name="transform_attributes",
        method_arg=ent_as_dict,
        inplace=True,
    ).values()
    if "sorted" in ent_as_dict.keys():
        sorted_ = ent_as_dict["sorted"]
    else:
        sorted_ = False
    cls_type_str = str(sel_cls.__fields__["type"].default)

    # Merge entries with the same name / uuid
    entities_copy = copy.deepcopy(ent)  # Copy to loop over
    del_keys = list()
    for ent_k, ent_v in entities_copy.items():
        if isinstance(ent_v, sel_cls) and ent_k not in del_keys:
            ent_uuid = ent_v.uuid
            for other_k, other_v in entities_copy.items():
                if (
                    isinstance(other_v, sel_cls)
                    and ent_k != other_k
                    and ent_uuid == other_v.uuid
                    and other_k not in del_keys
                ):
                    merge_res = merge_instances(ent_v, other_v)
                    # Overwrite with merge result
                    ent[ent_k] = merge_res
                    if sorted_:
                        if cls_type_str not in ent_as_dict.keys():
                            ent_as_dict[cls_type_str] = {ent_k: {"jsondata": {}}}
                        if ent_k not in ent_as_dict[cls_type_str].keys():
                            ent_as_dict[cls_type_str][ent_k] = {"jsondata": {}}
                        ent_as_dict[cls_type_str][ent_k]["jsondata"] = merge_res.dict()
                        del ent_as_dict[cls_type_str][other_k]
                    else:
                        ent_as_dict[ent_k]["jsondata"] = merge_res.dict()
                        del ent_as_dict[other_k]
                    del ent[other_k]
                    del_keys.append(other_k)
    print(f"Merge operation complete. Deleted keys: {del_keys}")
    return {"entities": ent, "entities_as_dict": ent_as_dict}


def get_uuid_from_object_via_type(obj: Any) -> Union[uuid_module.UUID, None]:
    """Get UUID from object via type. THis function assumes that the input object
    'obj' is either a dictionary or an instance of a OswBaseModel that specifies its
    type as a string or a list of strings of the full page title format."""
    # if isinstance(obj, model.OswBaseModel):
    #     type_ = getattr(obj, "type", None)
    if isinstance(obj, dict):
        type_ = obj.get("type", None)
    else:
        type_ = getattr(obj, "type", None)
    if type_ is not None:
        # if isinstance(type_, str):
        #     type_str = type_
        if isinstance(type_, list):
            type_str = str(type_[0])
        else:
            type_str = str(type_)
        match = re.match(
            pattern=REGEX_PATTERN["UUID from full page title"]["Pattern"],
            string=type_str,
        )
        uuid_str = match.group(3)
        return uuid_module.UUID(uuid_str)
    else:
        return None


def get_lang_specific_label(label: list, lang: str) -> Union[str, None]:
    """Get the label in a specific language from a list of labels"""
    for ele in label:
        if ele["lang"] == model.LangCode(lang):
            return ele["text"]
    return None


def get_coordinates(address):
    """Get the coordinates of an address, using the Nominatim API"""
    geolocator = Nominatim(user_agent="myapplication")
    location = geolocator.geocode(address)
    if location is None:
        return None, None
    return location.latitude, location.longitude


def get_one_or_more_element(inp: list):
    if len(inp) == 1:
        return inp[0]
    else:
        return inp


def unique_elements_or_none(seq: list):
    """Returns a list of the unique elements in 'seq' while preserving order.
    Taken from https://stackoverflow.com/a/480227
    """
    seen = set()
    seen_add = seen.add
    result = [x for x in seq if not (x in seen or seen_add(x))]
    if len(result) == 0:
        return None
    return result


def isclass(obj, cls):
    """Check if an object is an instance of a class or a subclass thereof.

    Parameters
    ----------
    obj
        Object to check.
    cls
        Class to check against.

    Returns
    -------
    result:
        True if obj is an instance of cls or a subclass thereof, False otherwise.
    """

    if isinstance(obj, cls):
        return True
    else:
        if isinstance(obj, dict):
            obj_type = obj.get("type")
        else:
            obj_type = getattr(obj, "type", None)
        cls_type = cls.__fields__["type"].default
        if isinstance(obj_type, list):
            obj_type.sort()
        if isinstance(cls_type, list):
            cls_type.sort()
        if obj_type == cls_type and obj_type is not None:
            return True
        return False


def flatten_list(nested_list):
    """Recursively flattens a nested list.

    Parameters
    ----------
    nested_list:
        Nested list to flatten.

    Returns
    -------
    flattened_list:
        Flattened list.

    Source
    ------
    https://www.geeksforgeeks.org/
    python-program-to-flatten-a-nested-list-using-recursion/
    """
    # check if list is empty
    if not (bool(nested_list)):
        return nested_list
    # to check instance of list is empty or not
    if isinstance(nested_list[0], list):
        # call function with sublist as argument
        return flatten_list(*nested_list[:1]) + flatten_list(nested_list[1:])
    # call function with sublist as argument
    flattened_list = nested_list[:1] + flatten_list(nested_list[1:])
    return flattened_list


def jsonpath_search_and_return_list_simple(
    jp_str: str, val_key: str, search_tar: dict
) -> Union[list, None]:
    jp_parse = jp.parse(path=jp_str)
    result = jp_parse.find(search_tar)
    list_ = []
    if result:
        for res in result:
            list_.append(res.value[val_key])
    return list_


def jsonpath_search_and_return_list(
    jp_str: str, val_key: str, search_tar: dict, class_to_match=None, warn: bool = True
) -> Union[list, None]:
    """

    Parameters
    ----------
    jp_str:
        The jsonpath string to search for
    val_key:
        The key of the value (of the search results) to be returned in a list
    search_tar:
        The dictionary to search in with jsonpath
    class_to_match:
        The class to match with the search results. Specifying this argument,
        reduces the search space dramatically and thus speeds up the search.
    warn:
        Whether to warn if the search return no results

    Returns
    -------
    result:
        List of values with key 'val_key', matching the jsonpath string 'jp_str'
        within the 'search_tar' dictionary

    Example
    -------
    Searching through entites of type HelperWikiFile within all entities (
    entities_as_dict) and returning the full_page_title of entities matching filename
    by the attribute name. Afterwards the attribute image is set to the result if it any
    >>> res = diu.jsonpath_search_and_return_list(
    >>>     jp_str=f'*[?name = "{filename}"]',
    >>>     val_key="full_page_title",
    >>>     search_tar=entities_as_dict,
    >>>     class_to_match=HelperWikiFile,
    >>> )
    >>> if len(res) > 0:
    >>>     self.image = res[0]

    """
    jp_parse = jp.parse(path=jp_str)
    if search_tar.get("sorted") is not None:
        sorted_ = search_tar["sorted"]
    else:
        sorted_ = False
    if sorted_ and class_to_match:
        # See definition in loop_and_call_method with argument 'sorted'
        try:
            cls_type = class_to_match.__fields__["type"].default
            # Search in a dramatically reduced number of entries
            result = jp_parse.find(search_tar[str(cls_type)])
        except Exception as e:
            warnings.warn(
                f"jsonpath_search_and_return_list() threw and " f"exception:\n{str(e)}"
            )
            result = jp_parse.find(search_tar)
    else:
        result = jp_parse.find(search_tar)
    list_ = []
    if result:
        for res in result:
            res_type = res.value["type"]
            res_type.sort()
            if class_to_match is None:
                list_.append(res.value[val_key])
            elif isclass(res.value, class_to_match):
                list_.append(res.value[val_key])
    if len(list_) == 0 and warn:
        warnings.warn(
            f"jsonpath_search_and_return_list() did not find any "
            f"results for the jsonpath string '{jp_str}'"
        )
    return list(set(flatten_list(list_)))


def camel_case(string, splitter: str = "\\s"):
    """Converts a string to camel case, by splitting it at spaces and capitalizing
    the first letter of each part, but the first.

    Parameters
    ----------
    string:
        String to convert to camel case.
    splitter:
        Regular expression string to split the string at. Default is any whitespace.

    Returns
    -------
    result:
        Camel case string.
    """
    if not string:
        return string
    parts = re.split(pattern=splitter, string=string)
    for i, part in enumerate(parts):
        parts[i] = part.capitalize()
    parts[0] = parts[0].lower()
    return "".join(parts)


def nan_empty_or_none(inp: Any) -> bool:
    """Checks if the input is None, an empty string, or a string "nan".

    Parameters
    ----------
    inp:
        Input to check.

    Returns
    -------
    result
    """
    if isinstance(inp, str):
        if inp == "nan" or inp == "":
            return True
        return False
    elif isinstance(inp, list):
        for ele in inp:
            if not nan_empty_or_none(ele):
                return False
        return True
    elif inp is None:
        return True
    else:
        try:
            if np.isnan(inp):
                return True
        except Exception as e:
            _ = e  # Bad catch
    return False


def regex_match_list(
    pattern: Union[str, dm.RegExPatternExtended], list_of_strings: List[str]
) -> List[Union[str, dm.MatchResult]]:
    """Returns a subset of the 'list_of_strings' that matched the regex 'pattern'.

    Parameters
    ----------
    pattern
    list_of_strings

    Returns
    -------
    matches:
        A lost of strings that matched the regex 'pattern'.
    """
    if isinstance(pattern, str) or isinstance(pattern, re.Pattern):
        matches = []
        for string in list_of_strings:
            if re.match(pattern=pattern, string=string):
                matches.append(string)
        return matches
    elif isinstance(pattern, dm.RegExPatternExtended):
        matches = []
        for string in list_of_strings:
            match_result_obj = pattern.match(string)
            if match_result_obj.match:
                matches.append(match_result_obj)
        return matches
    else:
        raise TypeError(
            f"regex_match_list() received an invalid type for the argument "
            f"'pattern': {type(pattern)}"
        )


def sort_dict_of_entities_as_dict_by_type(
    entities_as_dict: Dict[str, Dict],
    inplace: bool = True,
    sublevel_key: str = "jsondata",
) -> Dict:
    """

    Parameters
    ----------
    entities_as_dict:
        Dictionary of entities as dictionaries.
    inplace:
        If True, the input dictionary is sorted in place. If False, a copy is made.
    sublevel_key:
        The key of the sublevel in the entities_as_dict dictionary.

    Returns
    -------
    entities_as_dict:
        The sorted dictionary of entities as dictionaries.
    """
    if not inplace:
        entities_as_dict = copy.deepcopy(entities_as_dict)
    if "sorted" in entities_as_dict.keys():
        sorted_ = entities_as_dict["sorted"]
    else:
        sorted_ = False
    if not sorted_:  # Sort
        entities_as_dict_sorted = {}
        for key, val in entities_as_dict.items():
            val_type = str(val[sublevel_key]["type"])
            if val_type not in entities_as_dict_sorted.keys():
                entities_as_dict_sorted[val_type] = {}
            entities_as_dict_sorted[val_type][key] = val
        sorted_ = True
        entities_as_dict = entities_as_dict_sorted

    entities_as_dict["sorted"] = sorted_
    return entities_as_dict


def loop_and_call_method(
    entities: dict,
    method_name: str,
    method_arg: dict,
    inplace: bool = True,
    sort: bool = None,
    sublevel_key: str = "jsondata",
) -> dict:
    """Loops through dict 'entities', applies 'method' with 'method_arg' (also a
    dict, which is looped through) and either writes the result to the original dict
    ('inplace') or to a copy, which is then returned.

    Parameters
    ----------
    entities:
        Dictionary of entities.
    method_name:
        Name of the method to be called on each entity.
    method_arg:
        Positional argument passed to the method.
    inplace:
        If True, the method is applied to the original dict. If False, a copy is made.
    sort:
        If True, entities in 'method_arg' will be sorted according to attribute /
        dictionary key 'type' prio to being passed to the method. Returned
        'method_arg' will also contain the sorted structure.
    sublevel_key:
        String to be used on the dictionary sublevel.

    Returns
    -------
    result:
        Dictionary containing a dictionary of entities and a dictionary of entities
        as dict.
    """
    if sort is None:
        sort = ENABLE_SORTING
    if not inplace:
        entities = copy.deepcopy(entities)
        method_arg = copy.deepcopy(method_arg)
    if sort:
        method_arg = sort_dict_of_entities_as_dict_by_type(method_arg, inplace=True)
    # Loop and apply / call method
    entities_keys = copy.deepcopy(list(entities.keys()))
    for key in entities_keys:
        val = entities[key]
        method = val.__getattribute__(method_name)
        type_str = str(val.__getattribute__("type"))
        method_res = method(method_arg)
        if method_res:
            if key in entities.keys():
                entities[key] = val
            if sort and key in method_arg[type_str].keys():
                method_arg[type_str][key][sublevel_key] = val.dict()
            if not sort and key in method_arg.keys():
                method_arg[key][sublevel_key] = val.dict()
        else:
            if key in entities.keys():
                del entities[key]
            if sort and key in method_arg[type_str].keys():
                del method_arg[type_str][key]
            if not sort and key in method_arg.keys():
                del method_arg[key]
    return {"entities": entities, "method_arg": method_arg}


def merge_dicts(one: dict, other: dict) -> dict:
    """Merges two dictionaries, with the dictionary 'other' being the dominant
    dictionary, overwriting attributes of 'one'.
    """
    new_dict = dict()
    self_dict = one
    other_dict = other
    for key, self_val in self_dict.items():
        other_val = other_dict[key]
        if self_val == other_val:
            new_dict[key] = other_val
        else:
            if isinstance(self_val, list):
                self_set = set(self_val)
                other_set = set(other_val)
                in_other_but_not_in_self = other_set - self_set
                new_dict[key] = self_val + list(in_other_but_not_in_self)
            else:
                new_dict[key] = other_val
    return new_dict


def merge_instances(one, other, cls=None):
    """Merges two instances of any type (provided the class has a method 'dict()')
    with the instance 'other' being the dominant instance, overwriting attributes of
    'one'. The class is derived from 'one' if not specified in 'cls'. The merge is
    shallow, i.e. only the top level of the instances is merged.
    """
    if cls is None:
        cls = one.__class__
    new_dict = merge_dicts(one=one.dict(), other=other.dict())
    # todo: make deep merge available
    return cls(**new_dict)


def list_of_entities_to_dict_of_entities(
    list_of_entities: list,
    key_factory: Callable = None,
    key_factory_kwargs: dict = None,
    pass_entity_as_arg: bool = False,
) -> dict:
    """Converts a list of entities to a dictionary, with the keys being the
    'uuid' of the entities.

    Parameters
    ----------
    list_of_entities:
    key_factory:
        Function that takes kwargs as input and returns a key for the dictionary.
    key_factory_kwargs:
        Dictionary of keyword arguments to be passed to the key_factory function.
    pass_entity_as_arg:
        If True, the entity will be passed as an argument to the key_factory function.

    Returns
    -------
    dict_of_entities:
        Dictionary of entities.
    """
    dict_of_entities = {}
    for entity in list_of_entities:
        if getattr(entity, "uuid", None) is not None and key_factory is None:
            key = entity.uuid
        elif getattr(entity, "uuid", None) is None and key_factory is None:
            key = uuid_module.uuid4()
        elif key_factory is not None and key_factory_kwargs is None:
            if pass_entity_as_arg:
                key = key_factory(entity)
            else:
                key = key_factory()
        elif key_factory is not None and key_factory_kwargs is not None:
            if pass_entity_as_arg:
                key = key_factory(entity, **key_factory_kwargs)
            else:
                key = key_factory(**key_factory_kwargs)
        else:
            key = uuid_module.uuid4()
        dict_of_entities[str(key)] = entity
    return dict_of_entities


def insert_sublevel_to_dict(inp: dict, sublevel_key: str) -> dict:
    """Inserts a sublevel to a dictionary, with the key of the sublevel being
    'sublevel_key'."""
    result = {}
    for key, val in inp.items():
        result[key] = {sublevel_key: val}
    return result


def make_dict_jsonpath_compatible(inp: dict, sublevel_key: str = "jsondata") -> dict:
    """Makes a dictionary compatible with jsonpath, by inserting a sublevel with
    the key 'sublevel_key'."""
    result = {}
    for key, val in inp.items():
        # Try to access method 'dict' of the value, if it exists, use it, otherwise
        res = getattr(val, "dict", None)
        if res is not None:
            result[key] = {sublevel_key: val.dict()}
        else:
            result[key] = {sublevel_key: val}
    return result


def create_page_name_from_label(label: str) -> str:
    """Creates a page name from a label, by splitting the label at whitespaces,
    dashes, underscores, dots and slashes and capitalizing the first letter of
    each split."""
    splits = re.split(pattern="[\\s\\-\\_\\./]+", string=label)
    page_name = "".join([ele.capitalize() for ele in splits])
    return page_name


def get_entities_from_osw(
    category_to_search: Union[str, uuid_module.UUID],
    model_to_cast_to,
    credentials_fp,
    domain,
    debug: bool = False,
) -> list:
    """Gets entities from OSW, based on a category. The category can be specified by
    a string or a UUID. The entities are cast to the model specified by
    'model_to_cast_to'.

    Parameters
    ----------
    category_to_search:
        Category to search for.
    model_to_cast_to:
        Model to cast the entities to.
    credentials_fp:
        Filepath to the credentials file, used to access the OSW instance.
    debug:
        If True, prints debug information.

    Returns
    -------
    entities_from_osw:
        List of entities from OSW.
    """
    if isinstance(category_to_search, str):
        category_uuid = category_to_search.split("OSW")[-1]
    else:  # elif isinstance(category_to_search, uuid_module.UUID):
        category_uuid = str(category_to_search)
    cred_man = CredentialManager(cred_filepath=credentials_fp)
    osw_obj = OSW(site=WtSite(WtSite.WtSiteConfig(iri=domain, cred_mngr=cred_man)))
    wtsite_obj = osw_obj.site
    entities_from_osw = []
    if debug:
        print(f"Searching for instances of {category_to_search} in OSW...")
    entities = wtsite_obj.semantic_search(
        query=wt.SearchParam(
            query=f"[[HasType::Category:OSW{str(category_uuid).replace('-', '')}]]",
            debug=debug,
        )
    )
    for entity in entities:
        # entity = full page name
        page = wtsite_obj.get_page(WtSite.GetPageParam(titles=[entity])).pages[0]
        if page.exists:
            jsondata = page.get_slot_content("jsondata")
            jsondata["full_page_title"] = entity
            entities_from_osw.append(model_to_cast_to(**jsondata))
    return entities_from_osw


def uuid_to_full_page_title(
    uuid: Union[uuid_module.UUID, str],
    wiki_ns: str = "Item",
    prefix: str = "OSW",
    suffix: str = "",
) -> str:
    """Converts a UUID to a full page title, by prepending the wiki namespace and
    prefix to the UUID after removing dashes."""
    return f"{wiki_ns}:{prefix}{str(uuid).replace('-', '')}{suffix}"


def create_full_page_title(
    name: str, uuid_ns: uuid_module.UUID, wiki_ns: str = "Item", prefix: str = "OSW"
):
    """Creates a full page title from a name and a UUID namespace, by prepending the
    wiki namespace and prefix to the UUID after removing dashes."""
    _uuid = uuid_module.uuid5(uuid_ns, name)
    return uuid_to_full_page_title(uuid=_uuid, wiki_ns=wiki_ns, prefix=prefix)


def translate_list_with_deepl(
    seq: list, credentials_file_path: Union[str, Path] = None
) -> list:
    """Translates a list of strings with DeepL."""
    if credentials_file_path is None:
        credentials_file_path = CREDENTIALS_FILE_PATH_DEFAULT
    domains, accounts = wt.read_domains_from_credentials_file(credentials_file_path)
    domain = "api-free.deepl.com"
    auth = accounts[domain]["password"]
    translator = deepl.Translator(auth)
    return [translator.translate_text(ele, target_lang="EN-US").text for ele in seq]



from pathlib import Path

from osw.controller.page_package import PagePackageController as Package

# Provide information on the page package to be created
package = Package(
    # Package name
    name="OSW Core",
    # Package repository name - usually the GitHub repository name
    repo="world.opensemantic.core",
    # Package ID - usually the same as repo
    id="world.opensemantic.core",
    # Package subdirectory - usually resembling parts of the package name
    subdir="core",
    # Package branch - usually "main"
    branch="main",
    # (GitHub) Organization hosting the package repository
    repo_org="OpenSemanticWorld-Packages",
    # Provide a package description
    description=(
        "Provides core functionalities of OpenSemanticWorld / OpenSemanticLab"
    ),
    # Specify the package version - use semantic versioning
    version="0.2.1",
    # Author(s)
    author=["Simon Stier", "Lukas Gold"],
    # Publisher
    publisher="OpenSemanticWorld",
    # List the full page titles of the pages to be included in the package
    # You can include a comment in the same line, stating the page label
    page_titles=[
        "Module:Lustache",
        "Module:Lustache/Context",
        "Module:Lustache/Renderer",
        "Module:Lustache/Scanner",
        "Module:Entity",
        "Module:MwJson",
        "JsonSchema:Label",
        "JsonSchema:Description",
        "Category:Category",
        "Category:Entity",
        "Category:Item",
        "Template:Helper/UI/Tiles/Grid",
        "Template:Helper/UI/Tiles/Tile",
    ],
)
# Provide the information needed (only) to create the page package
config = Package.CreationConfig(
    # Specify the path to the credentials file
    credentials_file_path=Path(__file__).parent / "accounts.pwd.yaml",
    # Specify the domain of the OSW/OSL instance to load pages from
    domain="wiki-dev.open-semantic-lab.org",
    # Specify the path to the working directory - where the package is stored on disk
    working_dir=Path(__file__).parents[2] / "packages" / package.repo,
)
# Create the page package
package.create(creation_config=config)



import json
import os
from datetime import datetime
from pprint import pprint

import osw.wiki_tools as wt
from osw.wtsite import WtPage, WtSite

pwd_file_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "accounts.pwd.yaml"
)


def basic_text_manipulation():
    wikitext_org = """{{OslTemplate:LIMS/Device/Type
    |timestamp=2022-10-10T00:00:00.000Z
    |creator=C1;C2
    |display_name=Test Term
    |label=Device Test Type with Term
    |label_lang_code=en
    |description=Some description
    with line break
    |category=Category:OSWa444b0eeb79140d58a836a7fc6fc940a
    |relations={{OslTemplate:KB/Relation
    |property=IsRelatedTo
    |value=Term:OSW6b663c61c12d42e8be37d735dd2a869c
    }}SomeText{{OslTemplate:KB/Relation
    |property=IsRelatedTo
    |value=Term:OSW6b663c61c12d42e8be37d735dd2a869c
    }}
    }}
    =Details=
    some text

    {{Some/Template
    |p1=v1
    }}

    <br />
    {{OslTemplate:LIMS/Device/Type/Footer
    }}"""

    content_dict_1 = wt.create_flat_content_structure_from_wikitext(wikitext_org)
    pprint(content_dict_1)

    wikitext_2 = wt.get_wikitext_from_flat_content_structure(content_dict_1)
    print(wikitext_2)
    if wikitext_2 == wikitext_org:
        print("wikitext_2 == wikitext_org")
    else:
        print("wikitext_2 != wikitext_org")

    content_dict_3 = wt.create_flat_content_structure_from_wikitext(wikitext_2)
    pprint(content_dict_3)
    wikitext_3 = wt.get_wikitext_from_flat_content_structure(content_dict_3)
    print(wikitext_3)
    if wikitext_3 == wikitext_2:
        print("wikitext_3 == wikitext_2")
    else:
        print("wikitext_3 != wikitext_2")

    content_dict_3[0]["OslTemplate:LIMS/Device/Type"]["display_name"] = "NEW VALUE"
    print(wt.get_wikitext_from_flat_content_structure(content_dict_3))


def mass_page_edit():
    wtsite = WtSite.from_domain("wiki-dev.open-semantic-lab.org", pwd_file_path)
    # wtpage = wtsite.get_WtPage("LabNote:220601-sist-0001-ni")
    # wtpage = wtsite.get_WtPage("testesfesefsef")
    # wtpage.append_template("TestTemplate", {"p1": "v1"})
    # wtpage.append_text("Some text",)
    # wtpage.append_template("TestTemplate", {"p1": "v2"})

    # pprint(wtpage._dict)

    # res = wtpage.get_value("*.TestTemplate.p1")
    # pprint(res)
    # d = wtpage.set_value("*.TestTemplate.p1", "v3")

    # local_id = wtpage.title.split(":")[1]
    # wtpage.set_value("*.'OslTemplate:ELN/Entry/Header'", {"local_id" : [local_id]})

    # wtpage.update_content()
    # pprint(wtpage._dict)
    # wtpage.edit()
    # print(wtpage.changed)

    wtsite.modify_search_results(
        "semantic",
        "[[Category:labNote]]",
        lambda wtpage:
        # print("Lambda")#wtpage.title)
        wtpage.set_value(
            "*.'OslTemplate:ELN/Entry/Header'", {"id": [wtpage.title.split(":")[1]]}
        ).update_content(),
        limit=1,
        comment="[bot] set id from title",
        log=True,
        dryrun=True,
    )


def schema_renaming():
    wtsite = WtSite.from_domain("wiki-dev.open-semantic-lab.org", pwd_file_path)

    def modify(wtpage: WtPage):
        wtpage.content_replace("_osl_template", "osl_template")
        wtpage.content_replace("_osl_footer", "osl_footer")

    wtsite.modify_search_results(
        "prefix",
        "JsonSchema:",
        modify,
        limit=20,
        comment="rename keywords _osl* to osl*",
        log=True,
        dryrun=False,
    )


def slot_default_values():
    wtsite = WtSite.from_domain("wiki-dev.open-semantic-lab.org", pwd_file_path)

    def modify(wtpage: WtPage):
        edit = True
        # print(wtpage.get_last_changed_time())
        # if wtpage.get_last_changed_time() > datetime.fromisoformat('2023-01-15T12:00:00+00:00'):
        #    print("new")
        # else:
        #    print("old")
        #    edit = False
        if edit:
            if wtpage.get_slot_content("header") != "{{#invoke:Entity|header}}":
                wtpage.set_slot_content("header", "{{#invoke:Entity|header}}")
            if wtpage.get_slot_content("footer") != "{{#invoke:Entity|footer}}":
                wtpage.set_slot_content("footer", "{{#invoke:Entity|footer}}")

    wtsite.modify_search_results(
        "prefix",
        "Item:",
        # "semantic",
        # "[[IsA::Category:Category]]",
        modify,
        limit=300,
        comment="set default content of slot header and footer",
        log=True,
        dryrun=False,
    )


def text_replace():
    wtsite = WtSite.from_domain("wiki-dev.open-semantic-lab.org", pwd_file_path)

    def modify(wtpage: WtPage):
        edit = True
        if "#" in wtpage.title:
            edit = False
        if wtpage._page.redirect:
            edit = False
        if edit:
            for slot in wtpage._slots:
                content = wtpage.get_slot_content(slot)
                if isinstance(content, dict):
                    content = json.loads(json.dumps(content).replace("OSL", "OSW"))
                else:
                    content = content.replace("OSL", "OSW")
                wtpage.set_slot_content(slot, content)

    wtsite.modify_search_results(
        "semantic",
        # "[[IsA::Category:Category]]",
        "[[Item:+]]",
        # "prefix",
        # "Item:OSW",
        modify,
        limit=100,
        comment="replace OSL with OSW",
        log=True,
        dryrun=False,
    )


def move_page():
    wtsite = WtSite.from_domain("wiki-dev.open-semantic-lab.org", pwd_file_path)

    def modify(wtpage: WtPage):
        edit = True
        # print(wtpage.get_last_changed_time())
        if wtpage.get_last_changed_time() > datetime.fromisoformat(
            "2022-12-24T12:00:00+00:00"
        ):
            print("new")
        else:
            print("old")
            edit = False
        if "#" in wtpage.title:
            edit = False
        if wtpage._page.redirect:
            edit = False
        if edit:
            new_title = wtpage.title.replace("OSL", "OSW")
            wtpage.move(new_title=new_title, comment="replace OSL with OSW")

    wtsite.modify_search_results(
        "semantic",
        # "[[IsA::Category:Category]]",
        "[[~*File:OSL*]]",
        # "prefix",
        # "File:OSL",
        modify,
        limit=1000,
        comment="replace OSL with OSW",
        log=False,
        dryrun=True,
    )


# mass_page_edit()
# schema_renaming()
# slot_default_values()
# text_replace()
move_page()



from typing import Union
from uuid import UUID

import osw.model.entity as model


def get_osw_id(uuid: UUID) -> str:
    """Generates a OSW-ID based on the given uuid by prefixing "OSW" and removing
    all '-' from the uuid-string. Duplicates OSW.get_osw_id() from src/sw/core/osw.py

    Parameters
    ----------
    uuid
        uuid object, e. g. UUID("2ea5b605-c91f-4e5a-9559-3dff79fdd4a5")

    Returns
    -------
        OSW-ID string, e. g. OSW2ea5b605c91f4e5a95593dff79fdd4a5
    """
    return "OSW" + str(uuid).replace("-", "")


def get_uuid(osw_id) -> UUID:
    """Returns the uuid for a given OSW-ID. Duplicate of OSW.get_uuid() from src/sw/core/osw.py

    Parameters
    ----------
    osw_id
        OSW-ID string, e. g. OSW2ea5b605c91f4e5a95593dff79fdd4a5

    Returns
    -------
        uuid object, e. g. UUID("2ea5b605-c91f-4e5a-9559-3dff79fdd4a5")
    """
    return UUID(osw_id.replace("OSW", ""))


def get_namespace(entity: model.Entity) -> Union[str, None]:
    """determines the wiki namespace based on the entity's type/class

    Parameters
    ----------
    entity
        the entity to determine the namespace for

    Returns
    -------
        the namespace as a string or None if the namespace could not be determined
    """
    namespace = None

    if hasattr(entity, "meta") and entity.meta and entity.meta.wiki_page:
        if entity.meta.wiki_page.namespace:
            namespace = entity.meta.wiki_page.namespace

    #  (model classes may not exist => try except)
    if namespace is None:
        try:
            if issubclass(entity, model.Entity):
                namespace = "Category"
        except (TypeError, AttributeError):
            pass
    if namespace is None:
        try:
            if isinstance(entity, model.Item):
                namespace = "Item"
        except AttributeError:
            pass
    if namespace is None:
        try:
            if isinstance(entity, model.Category):
                namespace = "Category"
        except AttributeError:
            pass
    if namespace is None:
        try:
            if isinstance(entity, model.Property):
                namespace = "Property"
        except AttributeError:
            pass
    if namespace is None:
        try:
            if isinstance(entity, model.WikiFile):
                namespace = "File"
        except AttributeError:
            pass

    return namespace


def get_title(entity: model.Entity) -> Union[str, None]:
    """determines the wiki page title based on the entity's data

    Parameters
    ----------
    entity
        the entity to determine the title for

    Returns
    -------
        the title as a string or None if the title could not be determined
    """
    title = None

    if hasattr(entity, "meta") and entity.meta and entity.meta.wiki_page:
        if entity.meta.wiki_page.title:
            title = entity.meta.wiki_page.title

    if title is None:
        title = get_osw_id(entity.uuid)

    return title


def get_full_title(entity: model.Entity) -> Union[str, None]:
    """determines the wiki full title (namespace:title) based on the entity's data

    Parameters
    ----------
    entity
        the entity to determine the full title for

    Returns
    -------
        the full title as a string or None if the title could not be determined
    """
    namespace = get_namespace(entity)
    title = get_title(entity)
    if namespace is not None and title is not None:
        return namespace + ":" + title
    else:
        return title


def namespace_from_full_title(full_title: str) -> str:
    """extracts the namespace from a full title (namespace:title)

    Parameters
    ----------
    full_title
        the full title to extract the namespace from

    Returns
    -------
        the namespace as a string
    """
    return full_title.replace(title_from_full_title(full_title), "").replace(":", "")


def title_from_full_title(full_title: str) -> str:
    """extracts the title from a full title (namespace:title)

    Parameters
    ----------
    full_title
        the full title to extract the title from

    Returns
    -------
        the title as a string
    """
    return full_title.split(":")[-1]



from pprint import pprint

import osw.model.entity as model

my_entity = model.Item(
    label=[model.Label(text="MyItem")], statements=[model.Statement(predicate="IsA")]
)
pprint(my_entity.dict())
print(my_entity.json())



import os

import osw.model.entity as model
from osw.auth import CredentialManager
from osw.core import OSW
from osw.wtsite import WtSite

# create/update the password file under examples/accounts.pwd.yaml
pwd_file_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "accounts.pwd.yaml"
)
# pwd_file_path = "./accounts.pwd.yaml"
wtsite = WtSite(
    config=WtSite.WtSiteConfig(
        iri="wiki-dev.open-semantic-lab.org",
        cred_mngr=CredentialManager(cred_filepath=pwd_file_path),
    )
)
osw = OSW(site=wtsite)

# Load Tutorial Schema on demand
if not hasattr(model, "Tutorial"):
    osw.fetch_schema(
        OSW.FetchSchemaParam(
            schema_title="Category:OSW494f660e6a714a1a9681c517bbb975da", mode="replace"
        )
    )

# load instance HowToCreateANewArticle
title = "Item:OSW52c2c5a6bbc84fcb8eab0fa69857e7dc"
entity = osw.load_entity(title)
print(entity.__class__)
print(entity.label[0].text)  # we can access any attribute of model.Entity

tutorial_entity = entity.cast(model.Tutorial)  # explicit cast to model.Tutorial
print(
    tutorial_entity.required_predecessor
)  # we can access now any attribute of model.Tutorial

print(entity.json(exclude_none=True))  # export as json

# load all instances of Tutorial in parallel
tutorials = osw.query_instances(category="Category:OSW494f660e6a714a1a9681c517bbb975da")
print(tutorials)
osw.load_entity(tutorials)



import asyncio
import sys
from asyncio import Queue
from contextlib import redirect_stdout, suppress
from io import StringIO  # , BytesIO,

# import stdio_proxy
from pathlib import Path
from typing import IO, Callable, Dict, Iterable, List, Optional, Union

import dask
from dask.diagnostics import ProgressBar

# dask.config.set(scheduler="threads")
# with stdio_proxy.redirect_stdout(sys.stdout):
# "processes" scheduler leads to no messages ending up in the buffer / no flushing
# "threads" scheduler leads to all messages being printed immediatley without buffering
# with contextlib.redirect_stdout(sys.stdout):
# "processes" scheduler leads to no messages ending up in the buffer / no flushing
# "threads" scheduler leads to all buffered messages as expected


class MessageBuffer:
    """A class that buffers print messages and prints them all at once when
    flush is called. Can also be used as a context manager.

    Examples
    --------
    # Regular usage:
    # Create a BufferedPrint object
    >>> buffer = MessageBuffer()
    # Redirect print to the BufferedPrint object
    >>> print("Some message", file=buffer)
    # Flush the buffered messages and print them
    >>> buffer.flush()
    # As context manager:
    >>> with MessageBuffer() as buffer:
    ...     print("Some message", file=buffer)
    """

    def __init__(self, debug: bool = True):
        self.debug = debug
        self.messages = []
        self.closed = False  # Always open

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.debug:
            self.flush()
        self.closed = True

    def write(self, message):
        self.messages.append(message)

    def flush(self):
        # Process the buffered messages
        for message in self.messages:
            print(message, end="", file=sys.stdout)
        self.messages = []


class AsyncMessageBuffer(MessageBuffer):
    """Work in progress.


    A class that buffers print messages and prints them all at once when
    flush is called. Can also be used as a context manager.

    Examples
    --------
    # Regular usage:
    # Create a BufferedPrint object
    >>> buffer = MessageBuffer()
    # Redirect print to the BufferedPrint object
    >>> print("Some message", file=buffer)
    # Flush the buffered messages and print them
    >>> buffer.flush()
    # As context manager:
    >>> with MessageBuffer() as buffer:
    ...     print("Some message", file=buffer)
    """

    def __init__(self, debug: bool = True):
        super().__init__(debug=debug)
        self.messages = Queue()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.debug:
            self.flush()
        self.closed = True

    def write(self, message):
        self.messages.put(message)

    def flush(self):
        # Process the buffered messages
        while not self.messages.empty():
            message = self.messages.get()
            print(message, end="", file=sys.stdout)


def redirect_print(
    file_like: Union[IO, Path, MessageBuffer, AsyncMessageBuffer]
) -> Callable:
    def decorator(func_: Callable) -> Callable:
        """

        Parameters
        ----------
        func_:
            The function to decorate.

        Returns
        -------
        wrapper:
            The wrapped function.
        """

        def wrapper(*args, **kwargs):
            # First redirect stdout to a buffer
            with StringIO() as buf, redirect_stdout(buf):
                # with BytesIO() as buf, stdio_proxy.redirect_stdout(buf):
                return_val = func_(*args, **kwargs)
                output = buf.getvalue()
            # Reset stdout to the original value
            with suppress(AttributeError):
                # sys.stdout.close()
                sys.stdout = sys.__stdout__
                if sys.stdout.closed:
                    raise ValueError("sys.stdout is closed.")
            lines = output.splitlines()
            if getattr(file_like, "closed", None):
                raise ValueError(
                    "File or buffer is closed. Open prior to "
                    "execution of the decorated function."
                )
            if isinstance(file_like, Path) or isinstance(file_like, str):
                path = Path(file_like)
                if not path.exists():
                    raise FileNotFoundError(f"File {path} does not exist.")
                file_buffer = path.open("w")
                try:
                    for line in lines:
                        print(line, file=file_buffer)
                finally:
                    file_buffer.close()
            else:
                for line in lines:
                    print(line, file=file_like)

            return return_val

        return wrapper

    return decorator


def redirect_print_explicitly(
    func: Optional[Callable] = None,
    file_like: Optional[Union[IO, Path, MessageBuffer, AsyncMessageBuffer]] = None,
    line_print: Optional[Callable] = None,
) -> Callable:
    """A decorator that redirects print messages to a file or buffer and optionally
    applies a line_print function to each line. If line_print is not None, the print


    Parameters
    ----------
    func:
        The function to decorate.
    file_like:
        The file or buffer to redirect the print messages to.
    line_print:
        A function that will be called for each line of the print message.

    Returns
    -------
    Depending on usage, the decorated function or the decorator.

    Source
    ------
    Adapted from: https://bytepawn.com/python-decorators-for-data-scientists.html

    """

    # Code here is executed at decoration - when the decorator is used
    def decorator(func_: Callable, **kwargs) -> Callable:
        """

        Parameters
        ----------
        func_:
            The function to decorate.

        Returns
        -------
        wrapper:
            The wrapped function.
        """

        def wrapper(*args_, **kwargs_):
            kwargs_.update(kwargs)
            # First redirect stdout to a buffer
            with StringIO() as buf, redirect_stdout(buf):
                # with BytesIO() as buf, stdio_proxy.redirect_stdout(buf):
                return_val = func_(*args_, **kwargs_)
                output = buf.getvalue()
            # Reset stdout to the original value
            with suppress(AttributeError):
                # sys.stdout.close()
                sys.stdout = sys.__stdout__
                if sys.stdout.closed:
                    raise ValueError("sys.stdout is closed.")
            lines = output.splitlines()
            if line_print is not None:
                lines = [
                    line_print(line) for line in lines if line_print(line) is not None
                ]

            if getattr(file_like, "closed", None):
                raise ValueError(
                    "File or buffer is closed. Open prior to "
                    "execution of the decorated function."
                )
            if isinstance(file_like, Path) or isinstance(file_like, str):
                path = Path(file_like)
                if not path.exists():
                    raise FileNotFoundError(f"File {path} does not exist.")
                file_buffer = path.open("w")
                try:
                    for line in lines:
                        print(line, file=file_buffer)
                finally:
                    file_buffer.close()
            elif file_like is None:
                # todo: fix - does not store the messages until parallel execution
                #  finishes
                # Maybe because this code is only executed at the call of the wrapped
                # function and therefore, the buffer object is not shared between tasks
                buffer = MessageBuffer()
                for line in lines:
                    print(line, file=buffer)
                print("Printing buffered messages.")
                buffer.flush()
            else:
                for line in lines:
                    print(line, file=file_like)
                    # it seems that this write-operation is sometimes redirected in
                    # parallel execution as if print(line, file=None) was used

            return return_val

        return wrapper

    # Code here is executed at decoration - when the decorator is used
    if func is None:
        # decorator was used like @redirect(...)
        return decorator
    else:
        # decorator was used like @redirect, without parens
        return decorator(func_=func)


def parallelize(
    func: Callable,
    iterable: Iterable,
    flush_at_end: bool = False,
    progress_bar: bool = True,
    **kwargs,
):
    """A function to parallelize tasks with a progress bar and a message buffer.

    Parameters
    ----------
    func:
        The function to be executed in parallel.
    iterable:
        The iterable, who's items are to be passed as singles to the function.
    flush_at_end:
        If True, the message buffer will be printed (flushed) to stdout after parallel
        execution.
    progress_bar:
        If True, a progress bar will be displayed.
    kwargs:
        Keyword arguments to be passed to the function.z
    """
    with MessageBuffer(flush_at_end) as msg_buf:
        tasks = [
            dask.delayed(redirect_print_explicitly(func, msg_buf))(item, **kwargs)
            for item in iterable
        ]
        print(f"Performing parallel execution of {func.__name__} ({len(tasks)} tasks).")
        if progress_bar:
            with ProgressBar():
                results = dask.compute(*tasks)
        else:
            results = dask.compute(*tasks)
    return results


def list_files_and_directories(
    search_path: Union[str, Path], recursive: bool = True
) -> Dict[str, List[Path]]:
    """List files and subdirectories in the specified path. Return them as list
    nested in a directory.

    Parameters
    ----------
    search_path:
        The path to the directory.
    recursive:
        If True, the function will recursively search for files and directories.

    Returns
    -------
    result:
        A dictionary with a list each of files and directories in the directory.
    """
    search_path = Path(search_path)

    if not search_path.is_dir():
        raise ValueError(f"{search_path} is not a directory.")

    files = []
    directories = []

    for item in search_path.iterdir():
        if item.is_file():
            files.append(item)
        elif item.is_dir():
            directories.append(item)
            if recursive:
                ret_val = list_files_and_directories(item, recursive=True)
                files.extend(ret_val["files"])
                directories.extend(ret_val["directories"])

    return {"directories": directories, "files": files}


def file_in_paths(
    paths: List[Union[str, Path]], file_name: Union[str, Path]
) -> Dict[str, Union[bool, Path]]:
    result = False
    path = None
    for path in paths:
        if path.name == file_name:
            result = True
            break
    return {"found": result, "file path": path}


def async_parallelize(func: Callable, iterable: Iterable, **kwargs):
    """Work in progress"""

    async def inner(func_, iterable_, **kwargs_):
        async with AsyncMessageBuffer() as msg_buf:
            tasks = [
                dask.delayed(redirect_print_explicitly(func_, msg_buf))(item, **kwargs_)
                for item in iterable_
            ]
            print(
                f"Performing parallel execution of {func.__name__} "
                f"({len(tasks)} tasks)."
            )
            with ProgressBar():
                results = dask.compute(*tasks)
        return results

    return asyncio.run(inner(func, iterable, **kwargs))


class RedirectStdout(object):
    """Can be used as a decorator or context manager to temporarily redirect
    stdout to a buffer.

    Source
    ------
    Adapted in accordance to: https://gist.github.com/jasonbartz/8480004
    """

    def __init__(self, buffer: MessageBuffer, **kwargs):
        """Init and set vars"""
        self.buffer: MessageBuffer = buffer
        self.write_func = self.buffer.write
        self.kwargs = kwargs

    def __call__(self, function):
        """Call the function"""

        def wrapped_function(*args, **kwargs):
            with StringIO() as buf, redirect_stdout(buf):
                return_val = function(*args, **kwargs)
                out = buf.getvalue()
            with suppress(AttributeError):
                # sys.stdout.close()
                sys.stdout = sys.__stdout__
                if sys.stdout.closed:
                    raise ValueError("sys.stdout is closed.")
            self.write_func(out)
            return return_val

        return wrapped_function


# todo: fix
#  - If no message buffer is provided to file_like, the messages are not stored


# testing redirect
if __name__ == "__main__":
    from datetime import datetime
    from time import sleep

    print("Redirecting print messages to a buffer.")
    message_buffer = MessageBuffer()
    io_buffer = StringIO()
    file_as_buffer = open("xx.txt", "w")

    def prepend_timestamp(line):
        return f"{datetime.now().isoformat()}: {line}"

    def create_list_item(i, debug: bool = True):
        if debug:
            print(f"Creating item {i + 1}.")
        sleep(0.1)
        return i**2

    def create_list_dask(length: int, debug: bool = True) -> List[int]:
        message_buffer_ = MessageBuffer()
        # todo: line_print="line_numbers" doesn't return any message

        @redirect_print(
            file_like=message_buffer_,  # line_print=prepend_timestamp
        )
        def in_parallel(i_, length_, debug_: bool = True):
            if debug_:
                print(f"Creating item {i_ + 1} of {length_}")
            sleep(0.1)
            return i_**2

        tasks = []
        for i in range(length):
            tasks.append(dask.delayed(in_parallel)(i, length, debug))
        with ProgressBar():
            result = dask.compute(*tasks)
        sleep(0.1)
        if debug:
            print("Messages collected during parallel execution:")
            message_buffer_.flush()

        return result

    def create_list_dask_1(length: int, debug: bool = True) -> List[int]:
        @redirect_print
        def in_parallel(i_, length_, debug_: bool = True):
            if debug_:
                print(f"Creating item {i_ + 1} of {length_}")
            sleep(0.1)
            return i_**2

        tasks = []
        for i in range(length):
            tasks.append(dask.delayed(in_parallel)(i, length, debug))
        with ProgressBar():
            result = dask.compute(*tasks)

        return result

    def create_list_dask_2(length: int, debug: bool = True) -> List[int]:
        message_buffer_ = MessageBuffer()

        @RedirectStdout(buffer=message_buffer_)
        def in_parallel(i_, length_, debug_: bool = True):
            if debug_:
                print(f"Creating item {i_ + 1} of {length_}")
            sleep(0.1)
            return i_**2

        tasks = []
        for i in range(length):
            tasks.append(dask.delayed(in_parallel)(i, length, debug))
        with ProgressBar():
            result = dask.compute(*tasks)

        message_buffer_.flush()

        return result

    @redirect_print(file_like=message_buffer)
    def create_list(length: int) -> List[int]:
        print(f"Creating a list of length {length}")
        for i in range(length):
            sleep(0.1)
            print(f"Creating item {i + 1} of {length}")
        return [i for i in range(length)]

    @redirect_print
    def create_list_1(length: int) -> List[int]:
        print(f"Creating a list of length {length}")
        for i in range(length):
            sleep(0.1)
            print(f"Creating item {i + 1} of {length}")
        return [i for i in range(length)]

    @redirect_print_explicitly()
    def create_list_2(length: int) -> List[int]:
        print(f"Creating a list of length {length}")
        for i in range(length):
            sleep(0.1)
            print(f"Creating item {i + 1} of {length}")
        return [i for i in range(length)]

    @redirect_print(file_like=Path("xx.txt"))
    def create_list_3(length: int) -> List[int]:
        print(f"Creating a list of length {length}")
        for i in range(length):
            sleep(0.1)
            print(f"Creating item {i + 1} of {length}")
        return [i for i in range(length)]

    # @redirect_print(file_like=file_as_buffer)
    # def print_lines(num_lines):
    #     for i in range(num_lines):
    #         print(f'Line #{i + 1}')

    @redirect_print(file_like=message_buffer)
    def print_lines_1(num_lines):
        for i in range(num_lines):
            print(f"Line #{i + 1}")

    @redirect_print_explicitly(file_like=message_buffer, line_print=prepend_timestamp)
    def print_lines_2(num_lines):
        for i in range(num_lines):
            print(f"Line #{i + 1}")

    @redirect_print
    def print_lines_3(num_lines):
        for i in range(num_lines):
            print(f"Line #{i + 1}")

    print("Pre-execution")
    # print_lines(10)
    # print_lines_1(10)
    # print_lines_2(10)
    # print_lines_3(10)
    # test_list_dask = create_list_dask(10)
    # test_list_dask_1 = create_list_dask_1(10)
    # test_list_dask_2 = create_list_dask_2(10)
    test_list_new = parallelize(
        create_list_item,
        range(1000),
        flush_at_end=True,
    )
    # test_list_new = async_parallelize(create_list_item, range(10), debug=True)
    # test_list = create_list(10)
    # test_list_1 = create_list_1(10)
    # test_list_2 = create_list_2(10)
    # test_list_3 = create_list_3(10)
    print("Post-execution")
    print("Messages from the buffer:")
    message_buffer.flush()
    file_as_buffer.close()
    print("Messages from the file buffer:")
    with open("xx.txt", "r") as f:
        for line_ in f.readlines():
            print(line_, end="")
